# Copyright 2012-2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.customizations.cloudformation.deploy import DeployCommand
from awscli.customizations.cloudformation.package import PackageCommand


def initialize(cli):
    """
    The entry point for CloudFormation high level commands.
    """
    cli.register('building-command-table.cloudformation', inject_commands)


def inject_commands(command_table, session, **kwargs):
    """
    Called when the CloudFormation command table is being built. Used to
    inject new high level commands into the command list. These high level
    commands must not collide with existing low-level API call names.
    """
    command_table['package'] = PackageCommand(session)
    command_table['deploy'] = DeployCommand(session)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

# NOTE: This file is imported whenever a user hits TAB.  There should not
# be any expensive imports in this file.  If necessary, use lazy imports
# to ensure we only import heavyweight modules when we know we need them.
from awscli.autocomplete import db
from awscli.autocomplete.serverside import model, servercomp


def create_server_side_completer(index_filename, response_filter=None):
    return servercomp.ServerSideCompleter(
        model.DBCompletionLookup(db.DatabaseConnection(index_filename)),
        servercomp.LazyClientCreator(),
        response_filter,
    )



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import contextlib
import os
import sys
import tempfile
import zipfile
from datetime import datetime

from botocore.exceptions import ClientError

from awscli.compat import ZIP_COMPRESSION_MODE, BytesIO
from awscli.customizations.codedeploy.utils import validate_s3_location
from awscli.customizations.commands import BasicCommand
from awscli.customizations.exceptions import ParamValidationError

ONE_MB = 1 << 20
MULTIPART_LIMIT = 6 * ONE_MB


class Push(BasicCommand):
    NAME = 'push'

    DESCRIPTION = (
        'Bundles and uploads to Amazon Simple Storage Service (Amazon S3) an '
        'application revision, which is a zip archive file that contains '
        'deployable content and an accompanying Application Specification '
        'file (AppSpec file). If the upload is successful, a message is '
        'returned that describes how to call the create-deployment command to '
        'deploy the application revision from Amazon S3 to target Amazon '
        'Elastic Compute Cloud (Amazon EC2) instances.'
    )

    ARG_TABLE = [
        {
            'name': 'application-name',
            'synopsis': '--application-name <app-name>',
            'required': True,
            'help_text': (
                'Required. The name of the AWS CodeDeploy application to be '
                'associated with the application revision.'
            ),
        },
        {
            'name': 's3-location',
            'synopsis': '--s3-location s3://<bucket>/<key>',
            'required': True,
            'help_text': (
                r'Required. Information about the location of the application '
                r'revision to be uploaded to Amazon S3. You must specify both '
                r'a bucket and a key that represent the Amazon S3 bucket name '
                r'and the object key name. Content will be zipped before '
                r'uploading. Use the format s3://\<bucket\>/\<key\>'
            ),
        },
        {
            'name': 'ignore-hidden-files',
            'action': 'store_true',
            'default': False,
            'group_name': 'ignore-hidden-files',
            'help_text': (
                'Optional. Set the --ignore-hidden-files flag to not bundle '
                'and upload hidden files to Amazon S3; otherwise, set the '
                '--no-ignore-hidden-files flag (the default) to bundle and '
                'upload hidden files to Amazon S3.'
            ),
        },
        {
            'name': 'no-ignore-hidden-files',
            'action': 'store_true',
            'default': False,
            'group_name': 'ignore-hidden-files',
        },
        {
            'name': 'source',
            'synopsis': '--source <path>',
            'default': '.',
            'help_text': (
                'Optional. The location of the deployable content and the '
                'accompanying AppSpec file on the development machine to be '
                'zipped and uploaded to Amazon S3. If not specified, the '
                'current directory is used.'
            ),
        },
        {
            'name': 'description',
            'synopsis': '--description <description>',
            'help_text': (
                'Optional. A comment that summarizes the application '
                'revision. If not specified, the default string "Uploaded by '
                'AWS CLI \'time\' UTC" is used, where \'time\' is the current '
                'system time in Coordinated Universal Time (UTC).'
            ),
        },
    ]

    def _run_main(self, parsed_args, parsed_globals):
        self._validate_args(parsed_args)
        self.codedeploy = self._session.create_client(
            'codedeploy',
            region_name=parsed_globals.region,
            endpoint_url=parsed_globals.endpoint_url,
            verify=parsed_globals.verify_ssl,
        )
        self.s3 = self._session.create_client(
            's3', region_name=parsed_globals.region
        )
        self._push(parsed_args)
        return 0

    def _validate_args(self, parsed_args):
        validate_s3_location(parsed_args, 's3_location')
        if (
            parsed_args.ignore_hidden_files
            and parsed_args.no_ignore_hidden_files
        ):
            raise ParamValidationError(
                'You cannot specify both --ignore-hidden-files and '
                '--no-ignore-hidden-files.'
            )
        if not parsed_args.description:
            parsed_args.description = (
                f'Uploaded by AWS CLI {datetime.utcnow().isoformat()} UTC'
            )

    def _push(self, params):
        with self._compress(
            params.source, params.ignore_hidden_files
        ) as bundle:
            try:
                upload_response = self._upload_to_s3(params, bundle)
                params.eTag = upload_response['ETag'].replace('"', "")
                if 'VersionId' in upload_response:
                    params.version = upload_response['VersionId']
            except Exception as e:
                raise RuntimeError(
                    'Failed to upload \'%s\' to \'%s\': %s'
                    % (params.source, params.s3_location, str(e))
                )
        self._register_revision(params)

        if 'version' in params:
            version_string = f',version={params.version}'
        else:
            version_string = ''
        s3location_string = (
            f'--s3-location bucket={params.bucket},key={params.key},'
            f'bundleType=zip,eTag={params.eTag}{version_string}'
        )
        sys.stdout.write(
            'To deploy with this revision, run:\n'
            'aws deploy create-deployment '
            f'--application-name {params.application_name} {s3location_string} '
            '--deployment-group-name <deployment-group-name> '
            '--deployment-config-name <deployment-config-name> '
            '--description <description>\n'
        )

    @contextlib.contextmanager
    def _compress(self, source, ignore_hidden_files=False):
        source_path = os.path.abspath(source)
        appspec_path = os.path.sep.join([source_path, 'appspec.yml'])
        with tempfile.TemporaryFile('w+b') as tf:
            zf = zipfile.ZipFile(tf, 'w', allowZip64=True)
            # Using 'try'/'finally' instead of 'with' statement since ZipFile
            # does not have support context manager in Python 2.6.
            try:
                contains_appspec = False
                for root, dirs, files in os.walk(source, topdown=True):
                    if ignore_hidden_files:
                        files = [fn for fn in files if not fn.startswith('.')]
                        dirs[:] = [dn for dn in dirs if not dn.startswith('.')]
                    for fn in files:
                        filename = os.path.join(root, fn)
                        filename = os.path.abspath(filename)
                        arcname = filename[len(source_path) + 1 :]
                        if filename == appspec_path:
                            contains_appspec = True
                        zf.write(filename, arcname, ZIP_COMPRESSION_MODE)
                if not contains_appspec:
                    raise RuntimeError(f'{appspec_path} was not found')
            finally:
                zf.close()
            yield tf

    def _upload_to_s3(self, params, bundle):
        size_remaining = self._bundle_size(bundle)
        if size_remaining < MULTIPART_LIMIT:
            return self.s3.put_object(
                Bucket=params.bucket, Key=params.key, Body=bundle
            )
        else:
            return self._multipart_upload_to_s3(params, bundle, size_remaining)

    def _bundle_size(self, bundle):
        bundle.seek(0, 2)
        size = bundle.tell()
        bundle.seek(0)
        return size

    def _multipart_upload_to_s3(self, params, bundle, size_remaining):
        create_response = self.s3.create_multipart_upload(
            Bucket=params.bucket, Key=params.key
        )
        upload_id = create_response['UploadId']
        try:
            part_num = 1
            multipart_list = []
            bundle.seek(0)
            while size_remaining > 0:
                data = bundle.read(MULTIPART_LIMIT)
                upload_response = self.s3.upload_part(
                    Bucket=params.bucket,
                    Key=params.key,
                    UploadId=upload_id,
                    PartNumber=part_num,
                    Body=BytesIO(data),
                )
                multipart_list.append(
                    {'PartNumber': part_num, 'ETag': upload_response['ETag']}
                )
                part_num += 1
                size_remaining -= len(data)
            return self.s3.complete_multipart_upload(
                Bucket=params.bucket,
                Key=params.key,
                UploadId=upload_id,
                MultipartUpload={'Parts': multipart_list},
            )
        except ClientError as e:
            self.s3.abort_multipart_upload(
                Bucket=params.bucket, Key=params.key, UploadId=upload_id
            )
            raise e

    def _register_revision(self, params):
        revision = {
            'revisionType': 'S3',
            's3Location': {
                'bucket': params.bucket,
                'key': params.key,
                'bundleType': 'zip',
                'eTag': params.eTag,
            },
        }
        if 'version' in params:
            revision['s3Location']['version'] = params.version
        self.codedeploy.register_application_revision(
            applicationName=params.application_name,
            revision=revision,
            description=params.description,
        )



# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""
This customization adds the following scalar parameters to the
cloudwatch put-metric-data operation:

* --metric-name
* --dimensions
* --timestamp
* --value
* --statistic-values
* --unit
* --storage-resolution

"""

import decimal

from awscli.arguments import CustomArgument
from awscli.customizations.utils import validate_mutually_exclusive_handler
from awscli.utils import split_on_commas


def register_put_metric_data(event_handler):
    event_handler.register(
        'building-argument-table.cloudwatch.put-metric-data', _promote_args
    )
    event_handler.register(
        'operation-args-parsed.cloudwatch.put-metric-data',
        validate_mutually_exclusive_handler(
            ['metric_data'],
            [
                'metric_name',
                'timestamp',
                'unit',
                'value',
                'dimensions',
                'statistic_values',
            ],
        ),
    )


def _promote_args(argument_table, operation_model, **kwargs):
    # We're providing top level params for metric-data.  This means
    # that metric-data is now longer a required arg.  We do need
    # to check that either metric-data or the complex args we've added
    # have been provided.
    argument_table['metric-data'].required = False

    argument_table['metric-name'] = PutMetricArgument(
        'metric-name', help_text='The name of the metric.'
    )
    argument_table['timestamp'] = PutMetricArgument(
        'timestamp',
        help_text='The time stamp used for the metric.  '
        'If not specified, the default value is '
        'set to the time the metric data was '
        'received.',
    )
    argument_table['unit'] = PutMetricArgument(
        'unit', help_text='The unit of metric.'
    )
    argument_table['value'] = PutMetricArgument(
        'value',
        help_text='The value for the metric.  Although the --value '
        'parameter accepts numbers of type Double, '
        'Amazon CloudWatch truncates values with very '
        'large exponents.  Values with base-10 exponents '
        'greater than 126 (1 x 10^126) are truncated.  '
        'Likewise, values with base-10 exponents less '
        'than -130 (1 x 10^-130) are also truncated.',
    )

    argument_table['dimensions'] = PutMetricArgument(
        'dimensions',
        help_text=(
            'The --dimensions argument further expands '
            'on the identity of a metric using a Name=Value '
            'pair, separated by commas, for example: '
            '<code>--dimensions InstanceID=1-23456789,InstanceType=m1.small'
            '</code>. Note that the <code>--dimensions</code> argument has a '
            'different format when used in <code>get-metric-data</code>, '
            'where for the same example you would use the format '
            '<code>--dimensions Name=InstanceID,Value=i-aaba32d4 '
            'Name=InstanceType,value=m1.small </code>.'
        ),
    )
    argument_table['statistic-values'] = PutMetricArgument(
        'statistic-values',
        help_text='A set of statistical values describing the metric.',
    )

    metric_data = operation_model.input_shape.members['MetricData'].member
    storage_resolution = metric_data.members['StorageResolution']
    argument_table['storage-resolution'] = PutMetricArgument(
        'storage-resolution', help_text=storage_resolution.documentation
    )


def insert_first_element(name):
    def _wrap_add_to_params(func):
        def _add_to_params(self, parameters, value):
            if value is None:
                return
            if name not in parameters:
                # We're taking a shortcut here and assuming that the first
                # element is a struct type, hence the default value of
                # a dict.  If this was going to be more general we'd need
                # to have this paramterized, i.e. you pass in some sort of
                # factory function that creates the initial starting value.
                parameters[name] = [{}]
            first_element = parameters[name][0]
            return func(self, first_element, value)

        return _add_to_params

    return _wrap_add_to_params


class PutMetricArgument(CustomArgument):
    def add_to_params(self, parameters, value):
        method_name = '_add_param_%s' % self.name.replace('-', '_')
        return getattr(self, method_name)(parameters, value)

    @insert_first_element('MetricData')
    def _add_param_metric_name(self, first_element, value):
        first_element['MetricName'] = value

    @insert_first_element('MetricData')
    def _add_param_unit(self, first_element, value):
        first_element['Unit'] = value

    @insert_first_element('MetricData')
    def _add_param_timestamp(self, first_element, value):
        first_element['Timestamp'] = value

    @insert_first_element('MetricData')
    def _add_param_value(self, first_element, value):
        # Use a Decimal to avoid loss in precision.
        first_element['Value'] = decimal.Decimal(value)

    @insert_first_element('MetricData')
    def _add_param_dimensions(self, first_element, value):
        # Dimensions needs a little more processing.  We support
        # the key=value,key2=value syntax so we need to parse
        # that.
        dimensions = []
        for pair in split_on_commas(value):
            key, value = pair.split('=')
            dimensions.append({'Name': key, 'Value': value})
        first_element['Dimensions'] = dimensions

    @insert_first_element('MetricData')
    def _add_param_statistic_values(self, first_element, value):
        # StatisticValues is a struct type so we are parsing
        # a csv keyval list into a dict.
        statistics = {}
        for pair in split_on_commas(value):
            key, value = pair.split('=')
            # There are four supported values: Maximum, Minimum, SampleCount,
            # and Sum.  All of them are documented as a type double so we can
            # convert these to a decimal value to preserve precision.
            statistics[key] = decimal.Decimal(value)
        first_element['StatisticValues'] = statistics

    @insert_first_element('MetricData')
    def _add_param_storage_resolution(self, first_element, value):
        first_element['StorageResolution'] = int(value)



# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import argparse
import copy
import sys
from difflib import get_close_matches

HELP_BLURB = (
    "To see help text, you can run:\n"
    "\n"
    "  aws help\n"
    "  aws <command> help\n"
    "  aws <command> <subcommand> help\n"
)
USAGE = (
    "aws [options] <command> <subcommand> [<subcommand> ...] [parameters]\n"
    f"{HELP_BLURB}"
)


class ArgParseException(Exception):
    pass


class CommandAction(argparse.Action):
    """Custom action for CLI command arguments

    Allows the choices for the argument to be mutable. The choices
    are dynamically retrieved from the keys of the referenced command
    table
    """

    def __init__(self, option_strings, dest, command_table, **kwargs):
        self.command_table = command_table
        super().__init__(option_strings, dest, choices=self.choices, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values)

    @property
    def choices(self):
        return list(self.command_table.keys())

    @choices.setter
    def choices(self, val):
        # argparse.Action will always try to set this value upon
        # instantiation, but this value should be dynamically
        # generated from the command table keys. So make this a
        # NOOP if argparse.Action tries to set this value.
        pass


class CLIArgParser(argparse.ArgumentParser):
    Formatter = argparse.RawTextHelpFormatter

    def _check_value(self, action, value):
        """
        It's probably not a great idea to override a "hidden" method
        but the default behavior is pretty ugly and there doesn't
        seem to be any other way to change it.
        """
        # converted value must be one of the choices (if specified)
        if action.choices is not None and value not in action.choices:
            msg = [f"Found invalid choice '{value}'\n"]
            possible = get_close_matches(value, action.choices, cutoff=0.8)
            if possible:
                extra = ['Maybe you meant:\n']
                for word in possible:
                    extra.append(f'  * {word}')
                msg.extend(extra)
            raise argparse.ArgumentError(action, '\n'.join(msg))

    def parse_known_args(self, args, namespace=None):
        parsed, remaining = super().parse_known_args(args, namespace)
        terminal_encoding = getattr(sys.stdin, 'encoding', 'utf-8')
        if terminal_encoding is None:
            # In some cases, sys.stdin won't have an encoding set,
            # (e.g if it's set to a StringIO).  In this case we just
            # default to utf-8.
            terminal_encoding = 'utf-8'
        for arg, value in vars(parsed).items():
            if isinstance(value, bytes):
                setattr(parsed, arg, value.decode(terminal_encoding))
            elif isinstance(value, list):
                encoded = []
                for v in value:
                    if isinstance(v, bytes):
                        encoded.append(v.decode(terminal_encoding))
                    else:
                        encoded.append(v)
                setattr(parsed, arg, encoded)
        return parsed, remaining

    def error(self, message):
        """error(message: string)

        NOTE: This is lifted directly from argparse, the only difference being
        we use 252 for parsing errors.

        Raises exception with a usage message incorporating the message.

        If you override this in a subclass, it should not return -- it
        should raise an exception.
        """
        usage_message = self.format_usage()
        raise ArgParseException(f'{message}\n\n{usage_message}')


class MainArgParser(CLIArgParser):
    Formatter = argparse.RawTextHelpFormatter

    def __init__(
        self,
        command_table,
        version_string,
        description,
        argument_table,
        prog=None,
    ):
        super().__init__(
            formatter_class=self.Formatter,
            add_help=False,
            conflict_handler='resolve',
            description=description,
            usage=USAGE,
            prog=prog,
        )
        self._build(command_table, version_string, argument_table)

    def _create_choice_help(self, choices):
        help_str = ''
        for choice in sorted(choices):
            help_str += f'* {choice}\n'
        return help_str

    def _build(self, command_table, version_string, argument_table):
        for argument_name in argument_table:
            argument = argument_table[argument_name]
            argument.add_to_parser(self)
        self.add_argument(
            '--version',
            action="version",
            version=version_string,
            help='Display the version of this tool',
        )
        self.add_argument(
            'command', action=CommandAction, command_table=command_table
        )


class ServiceArgParser(CLIArgParser):
    def __init__(self, operations_table, service_name):
        super().__init__(
            formatter_class=argparse.RawTextHelpFormatter,
            add_help=False,
            conflict_handler='resolve',
            usage=USAGE,
        )
        self._build(operations_table)
        self._service_name = service_name

    def _build(self, operations_table):
        self.add_argument(
            'operation', action=CommandAction, command_table=operations_table
        )


class ArgTableArgParser(CLIArgParser):
    """CLI arg parser based on an argument table."""

    def __init__(self, argument_table, command_table=None):
        # command_table is an optional subcommand_table.  If it's passed
        # in, then we'll update the argparse to parse a 'subcommand' argument
        # and populate the choices field with the command table keys.
        super().__init__(
            formatter_class=self.Formatter,
            add_help=False,
            usage=USAGE,
            conflict_handler='resolve',
        )
        if command_table is None:
            command_table = {}
        self._build(argument_table, command_table)

    def _build(self, argument_table, command_table):
        for arg_name in argument_table:
            argument = argument_table[arg_name]
            argument.add_to_parser(self)
        if command_table:
            self.add_argument(
                'subcommand',
                action=CommandAction,
                command_table=command_table,
                nargs='?',
            )

    def parse_known_args(self, args, namespace=None):
        if len(args) == 1 and args[0] == 'help':
            namespace = argparse.Namespace()
            namespace.help = 'help'
            return namespace, []
        else:
            return super().parse_known_args(args, namespace)


class SubCommandArgParser(ArgTableArgParser):
    """Parse args for a subcommand but do not consume the arg table.

    This is similar to the ArgTableArgParser, except it doesn't consume
    any args from the provided arg table, though it will respect the
    arg table when looking for a subcommand.

    """

    def parse_known_args(self, args, namespace=None):
        parsed_args, remaining = super().parse_known_args(args, namespace)
        if getattr(parsed_args, 'subcommand', None) is not None:
            new_args = self._remove_subcommand(args, parsed_args)
            return new_args, parsed_args.subcommand
        return None

    def _remove_subcommand(self, args, parsed_args):
        # We want to remove only the subcommand from the args list
        # and keep everything else.  We should be safe to remove
        # the first argument that matches the subcommand name.
        #
        # .parse_known_args() assumes that any args that it doesn't
        # understand do not consume any values.  For example:
        #
        # aws ecs --cluster mycluster describe-tasks --tasks foo
        #
        # is not a valid command, because we ignore `--cluster`
        # and assume that `mycluster` is a subcommand.
        #
        # However, this command works:
        #
        # aws ecs --cluster=mycluster desribe-tasks --tasks foo
        #
        # Therefore we don't have to worry about the case where
        # a param *value* shadows the parsed subcommand because
        # the existing parser would have already errored out.
        new_args = args[:]
        new_args.remove(parsed_args.subcommand)
        return new_args

    def _build(self, argument_table, command_table):
        # In order to check for a subcommand while still respecting
        # the arg table, we do need to consider the arg table, but not
        # fail if any of the required args aren't provided.  We don't
        # want to mutate the arg table that's provided to us, so we
        # make a copy of it and then set all the required to not required.
        non_required_arg_table = self._non_required_arg_table(argument_table)
        for arg_name in non_required_arg_table:
            argument = non_required_arg_table[arg_name]
            argument.add_to_parser(self)
        if command_table:
            self.add_argument(
                'subcommand',
                action=CommandAction,
                command_table=command_table,
                nargs='?',
            )

    def _non_required_arg_table(self, argument_table):
        arg_table_copy = {}
        for key, value in argument_table.items():
            copied = copy.copy(value)
            copied.required = False
            arg_table_copy[key] = copied
        return arg_table_copy


class FirstPassGlobalArgParser(CLIArgParser):
    def __init__(self, *args, **kwargs):
        kwargs['add_help'] = False
        super().__init__(*args, **kwargs)
        self._build()

    def _build(self):
        self.add_argument('--profile', type=str)
        self.add_argument('--debug', action='store_true', default=False)



# Copyright 2022 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import csv
import io
import json
import os
import sys
from collections import namedtuple
from datetime import datetime

from awscli.customizations.commands import BasicCommand
from awscli.customizations.exceptions import ConfigurationError

# Takes botocore's ReadOnlyCredentials and exposes an expiry_time.
Credentials = namedtuple(
    'Credentials', ['access_key', 'secret_key', 'token', 'expiry_time']
)


def convert_botocore_credentials(credentials):
    # Converts botocore credentials to our `Credentials` type.
    frozen = credentials.get_frozen_credentials()
    expiry_time_str = None
    # Botocore does not expose an attribute for the expiry_time of temporary
    # credentials, so for the time being we need to access an internal
    # attribute to retrieve this info.  We're following up to see if botocore
    # can make this a public attribute.
    expiry_time = getattr(credentials, '_expiry_time', None)
    if expiry_time is not None and isinstance(expiry_time, datetime):
        expiry_time_str = expiry_time.isoformat()
    return Credentials(
        access_key=frozen.access_key,
        secret_key=frozen.secret_key,
        token=frozen.token,
        expiry_time=expiry_time_str,
    )


class BaseCredentialFormatter:
    FORMAT = None
    DOCUMENTATION = ""

    def __init__(self, stream=None):
        if stream is None:
            stream = sys.stdout
        self._stream = stream

    def display_credentials(self, credentials):
        pass


class BasePerLineFormatter(BaseCredentialFormatter):
    _VAR_FORMAT = 'export {var_name}={var_value}'

    def display_credentials(self, credentials):
        output = self._format_line(
            'AWS_ACCESS_KEY_ID', credentials.access_key
        ) + self._format_line('AWS_SECRET_ACCESS_KEY', credentials.secret_key)
        if credentials.token is not None:
            output += self._format_line('AWS_SESSION_TOKEN', credentials.token)
        if credentials.expiry_time is not None:
            output += self._format_line(
                'AWS_CREDENTIAL_EXPIRATION', credentials.expiry_time
            )
        self._stream.write(output)

    def _format_line(self, var_name, var_value):
        return (
            self._VAR_FORMAT.format(var_name=var_name, var_value=var_value)
            + '\n'
        )


class BashEnvVarFormatter(BasePerLineFormatter):
    FORMAT = 'env'
    DOCUMENTATION = (
        "Display credentials as exported shell variables: "
        "``export AWS_ACCESS_KEY_ID=EXAMPLE``"
    )
    _VAR_FORMAT = 'export {var_name}={var_value}'


class BashNoExportEnvFormatter(BasePerLineFormatter):
    FORMAT = 'env-no-export'
    DOCUMENTATION = (
        "Display credentials as non-exported shell variables: "
        "``AWS_ACCESS_KEY_ID=EXAMPLE``"
    )
    _VAR_FORMAT = '{var_name}={var_value}'


class PowershellFormatter(BasePerLineFormatter):
    FORMAT = 'powershell'
    DOCUMENTATION = (
        'Display credentials as PowerShell environment variables: '
        '``$Env:AWS_ACCESS_KEY_ID="EXAMPLE"``'
    )
    _VAR_FORMAT = '$Env:{var_name}="{var_value}"'


class WindowsCmdFormatter(BasePerLineFormatter):
    FORMAT = 'windows-cmd'
    DOCUMENTATION = (
        'Display credentials as Windows cmd environment variables: '
        '``set AWS_ACCESS_KEY_ID=EXAMPLE``'
    )
    _VAR_FORMAT = 'set {var_name}={var_value}'


class FishShellFormatter(BasePerLineFormatter):
    FORMAT = 'fish'
    DOCUMENTATION = (
        'Display credentials as Fish shell environment variables: '
        '``set -gx AWS_ACCESS_KEY_ID "EXAMPLE"``'
    )
    _VAR_FORMAT = 'set -gx {var_name} "{var_value}"'


class CredentialProcessFormatter(BaseCredentialFormatter):
    FORMAT = 'process'
    DOCUMENTATION = (
        "Display credentials as JSON output, in the schema "
        "expected by the ``credential_process`` config value."
        "This enables any library or tool that supports "
        "``credential_process`` to use the AWS CLI's credential "
        "resolution process: ``credential_process = aws configure "
        "export-credentials --profile myprofile``"
    )

    def display_credentials(self, credentials):
        output = {
            'Version': 1,
            'AccessKeyId': credentials.access_key,
            'SecretAccessKey': credentials.secret_key,
        }
        if credentials.token is not None:
            output['SessionToken'] = credentials.token
        if credentials.expiry_time is not None:
            output['Expiration'] = credentials.expiry_time
        self._stream.write(
            json.dumps(output, indent=2, separators=(',', ': '))
        )
        self._stream.write('\n')


SUPPORTED_FORMATS = {
    format_cls.FORMAT: format_cls
    for format_cls in [
        CredentialProcessFormatter,
        BashEnvVarFormatter,
        BashNoExportEnvFormatter,
        PowershellFormatter,
        WindowsCmdFormatter,
        FishShellFormatter,
    ]
}


def generate_docs(formats):
    lines = [
        'The output format to display credentials.  '
        'Defaults to `process`.  ',
        '<ul>',
    ]
    for name, cls in formats.items():
        line = f'<li>``{name}`` - {cls.DOCUMENTATION} </li>'
        lines.append(line)
    lines.append('</ul>')
    return '\n'.join(lines)


class ConfigureExportCredentialsCommand(BasicCommand):
    NAME = 'export-credentials'
    SYNOPSIS = 'aws configure export-credentials --profile profile-name'
    DESCRIPTION = (
        "Export credentials in various formats.  This command will retrieve "
        "AWS credentials using the AWS CLI's credential resolution process "
        "and display the credentials in the specified ``--format``. By "
        "default, the output format is ``process``, which is a JSON format "
        "that's expected by the credential process feature supported by the "
        "AWS SDKs and Tools.  This command ignores the global ``--query`` and "
        "``--output`` options."
    )
    ARG_TABLE = [
        {
            'name': 'format',
            'help_text': generate_docs(SUPPORTED_FORMATS),
            'action': 'store',
            'choices': list(SUPPORTED_FORMATS),
            'default': CredentialProcessFormatter.FORMAT,
        },
    ]
    _RECURSION_VAR = '_AWS_CLI_PROFILE_CHAIN'
    # Two levels is reasonable because you might explicitly run
    # "aws configure export-credentials" with a profile that is configured
    # with a credential_process of "aws configure export-credentials".
    # So we'll give one more level of recursion for padding and then
    # error out when we hit _MAX_RECURSION.
    _MAX_RECURSION = 4

    def __init__(self, session, out_stream=None, error_stream=None, env=None):
        super(ConfigureExportCredentialsCommand, self).__init__(session)
        if out_stream is None:
            out_stream = sys.stdout
        if error_stream is None:
            error_stream = sys.stderr
        if env is None:
            env = os.environ
        self._out_stream = out_stream
        self._error_stream = error_stream
        self._env = env

    def _detect_recursion_barrier(self):
        profile = self._get_current_profile()
        seen_profiles = self._parse_profile_chain(
            self._env.get(self._RECURSION_VAR, '')
        )
        if len(seen_profiles) >= self._MAX_RECURSION:
            raise ConfigurationError(
                f"Maximum recursive credential process resolution reached "
                f"({self._MAX_RECURSION}).\n"
                f"Profiles seen: {' -> '.join(seen_profiles)}"
            )
        if profile in seen_profiles:
            raise ConfigurationError(
                f"Credential process resolution detected an infinite loop, "
                f"profile cycle: {' -> '.join(seen_profiles + [profile])}\n"
            )

    def _update_recursion_barrier(self):
        profile = self._get_current_profile()
        seen_profiles = self._parse_profile_chain(
            self._env.get(self._RECURSION_VAR, '')
        )
        seen_profiles.append(profile)
        serialized = self._serialize_to_csv_str(seen_profiles)
        self._env[self._RECURSION_VAR] = serialized

    def _serialize_to_csv_str(self, profiles):
        out = io.StringIO()
        w = csv.writer(out)
        w.writerow(profiles)
        serialized = out.getvalue().strip()
        return serialized

    def _get_current_profile(self):
        profile = self._session.get_config_variable('profile')
        if profile is None:
            profile = 'default'
        return profile

    def _parse_profile_chain(self, value):
        result = list(csv.reader([value]))[0]
        return result

    def _run_main(self, parsed_args, parsed_globals):
        self._detect_recursion_barrier()
        self._update_recursion_barrier()
        try:
            creds = self._session.get_credentials()
        except Exception as e:
            original_msg = str(e).strip()
            raise ConfigurationError(
                f"Unable to retrieve credentials: {original_msg}\n"
            )
        if creds is None:
            raise ConfigurationError(
                "Unable to retrieve credentials: no credentials found"
            )
        creds_with_expiry = convert_botocore_credentials(creds)
        formatter = SUPPORTED_FORMATS[parsed_args.format](self._out_stream)
        formatter.display_credentials(creds_with_expiry)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import ruamel.yaml
from botocore.compat import OrderedDict

from awscli.utils import dump_yaml_to_str


def _ordered_constructor(loader, node):
    loader.flatten_mapping(node)
    return OrderedDict(loader.construct_pairs(node))


def _ordered_representer(dumper, data):
    return dumper.represent_mapping(
        ruamel.yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, data.items()
    )


def ordered_yaml_load(stream):
    """Load an OrderedDict object from a yaml stream."""
    yaml = ruamel.yaml.YAML(typ="safe", pure=True)
    yaml.Constructor.add_constructor(
        ruamel.yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
        _ordered_constructor,
    )

    return yaml.load(stream)


def ordered_yaml_dump(to_dump, stream=None):
    """
    Dump an OrderedDict object to yaml.

    :param to_dump: The OrderedDict to dump
    :type to_dump: OrderedDict

    :param stream: The file to dump to
    If not given or if None, only return the value
    :type stream: file
    """
    yaml = ruamel.yaml.YAML(typ="safe", pure=True)
    yaml.default_flow_style = False
    yaml.width = float('inf')
    yaml.Representer.add_representer(OrderedDict, _ordered_representer)

    if stream is None:
        return dump_yaml_to_str(yaml, to_dump)

    yaml.dump(to_dump, stream)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Add S3 specific event streaming output arg."""

from awscli.arguments import CustomArgument

STREAM_HELP_TEXT = 'Filename where the records will be saved'


class DocSectionNotFoundError(Exception):
    pass


def register_event_stream_arg(event_handlers):
    event_handlers.register(
        'building-argument-table.s3api.select-object-content',
        add_event_stream_output_arg,
    )

    event_handlers.register_last(
        'doc-output.s3api.select-object-content', replace_event_stream_docs
    )


def register_document_expires_string(event_handlers):
    event_handlers.register_last('doc-output.s3api', document_expires_string)


def add_event_stream_output_arg(
    argument_table, operation_model, session, **kwargs
):
    argument_table['outfile'] = S3SelectStreamOutputArgument(
        name='outfile',
        help_text=STREAM_HELP_TEXT,
        cli_type_name='string',
        positional_arg=True,
        stream_key=operation_model.output_shape.serialization['payload'],
        session=session,
    )


def replace_event_stream_docs(help_command, **kwargs):
    doc = help_command.doc
    current = ''
    while current != '======\nOutput\n======':
        try:
            current = doc.pop_write()
        except IndexError:
            # This should never happen, but in the rare case that it does
            # we should be raising something with a helpful error message.
            raise DocSectionNotFoundError(
                'Could not find the "output" section for the command: %s'
                % help_command
            )
    doc.write('======\nOutput\n======\n')
    doc.write(
        "This command generates no output.  The selected "
        "object content is written to the specified outfile.\n"
    )


def document_expires_string(help_command, **kwargs):
    doc = help_command.doc
    expires_field_idx = doc.find_last_write('Expires -> (timestamp)')

    if expires_field_idx is None:
        return

    deprecation_note_and_expires_string = [
        f'\n\n\n{" " * doc.style.indentation * doc.style.indent_width}',
        '.. note::',
        f'\n\n\n{" " * (doc.style.indentation + 1) * doc.style.indent_width}',
        'This member has been deprecated. Please use `ExpiresString` instead.\n',
        f'\n\n{" " * doc.style.indentation * doc.style.indent_width}',
        f'\n\n{" " * doc.style.indentation * doc.style.indent_width}',
        'ExpiresString -> (string)\n\n',
        '\tThe raw, unparsed value of the ``Expires`` field.',
        f'\n\n{" " * doc.style.indentation * doc.style.indent_width}',
    ]

    for idx, write in enumerate(deprecation_note_and_expires_string):
        # We add 4 to the index of the expires field name because each
        # field in the output section consists of exactly 4 elements.
        doc.insert_write(expires_field_idx + idx + 4, write)


class S3SelectStreamOutputArgument(CustomArgument):
    _DOCUMENT_AS_REQUIRED = True

    def __init__(self, stream_key, session, **kwargs):
        super(S3SelectStreamOutputArgument, self).__init__(**kwargs)
        # This is the key in the response body where we can find the
        # streamed contents.
        self._stream_key = stream_key
        self._output_file = None
        self._session = session

    def add_to_params(self, parameters, value):
        self._output_file = value
        self._session.register(
            'after-call.s3.SelectObjectContent', self.save_file
        )

    def save_file(self, parsed, **kwargs):
        # This method is hooked into after-call which fires
        # before the error checking happens in the client.
        # Therefore if the stream_key is not in the parsed
        # response we immediately return and let the default
        # error handling happen.
        if self._stream_key not in parsed:
            return
        event_stream = parsed[self._stream_key]
        with open(self._output_file, 'wb') as fp:
            for event in event_stream:
                if 'Records' in event:
                    fp.write(event['Records']['Payload'])
        # We don't want to include the streaming param in
        # the returned response, it's not JSON serializable.
        del parsed[self._stream_key]



# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json
import logging
import os

from botocore.exceptions import ClientError

from awscli.customizations.commands import BasicCommand
from awscli.customizations.sso.utils import AWS_CREDS_CACHE_DIR, SSO_TOKEN_DIR

LOG = logging.getLogger(__name__)


class LogoutCommand(BasicCommand):
    NAME = 'logout'
    DESCRIPTION = (
        'Removes all cached AWS IAM Identity Center access tokens and any '
        'cached temporary AWS credentials retrieved with SSO access tokens '
        'across all profiles. To use these profiles again, run: '
        '``aws sso login``'
    )
    ARG_TABLE = []

    def _run_main(self, parsed_args, parsed_globals):
        SSOTokenSweeper(self._session, parsed_globals).delete_credentials(
            SSO_TOKEN_DIR
        )
        SSOCredentialSweeper().delete_credentials(AWS_CREDS_CACHE_DIR)
        return 0


class BaseCredentialSweeper:
    def delete_credentials(self, creds_dir):
        if not os.path.isdir(creds_dir):
            return
        filenames = os.listdir(creds_dir)
        for filename in filenames:
            filepath = os.path.join(creds_dir, filename)
            contents = self._get_json_contents(filepath)
            if contents is None:
                continue
            if self._should_delete(contents):
                self._before_deletion(contents)
                os.remove(filepath)

    def _should_delete(self, filename):
        raise NotImplementedError('_should_delete')

    def _get_json_contents(self, filename):
        try:
            with open(filename) as f:
                return json.load(f)
        except Exception:
            # We do not want to include the traceback in the exception
            # so that we do not accidentally log sensitive contents because
            # of the exception or its Traceback.
            LOG.debug('Failed to load: %s', filename)
            return None

    def _before_deletion(self, contents):
        pass


class SSOTokenSweeper(BaseCredentialSweeper):
    def __init__(self, session, parsed_globals):
        self._session = session
        self._parsed_globals = parsed_globals

    def _should_delete(self, contents):
        return 'accessToken' in contents

    def _before_deletion(self, contents):
        # If the sso region is present in the cached token, construct a client
        # and invoke the logout api to invalidate the token before deleting it.
        sso_region = contents.get('region')
        if sso_region:
            sso = self._session.create_client(
                'sso',
                region_name=sso_region,
                verify=self._parsed_globals.verify_ssl,
            )
            try:
                sso.logout(accessToken=contents['accessToken'])
            except ClientError:
                # The token may already be expired or otherwise invalid. If we
                # get a client error on logout just log and continue on
                LOG.debug('Failed to call logout API:', exc_info=True)


class SSOCredentialSweeper(BaseCredentialSweeper):
    def _should_delete(self, contents):
        return contents.get('ProviderType') == 'sso'



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import errno
import logging
import os
import time

from botocore.utils import percent_encode_sequence
from s3transfer.subscribers import BaseSubscriber

from awscli.customizations.s3 import utils

LOGGER = logging.getLogger(__name__)


class CreateDirectoryError(Exception):
    pass


# TODO: Eventually port this down to the BaseSubscriber or a new subscriber
# class in s3transfer. The functionality is very convenient but may need
# some further design decisions to make it a feature in s3transfer.
class OnDoneFilteredSubscriber(BaseSubscriber):
    """Subscriber that differentiates between successes and failures

    It is really a convenience class so developers do not have to have
    to constantly remember to have a general try/except around future.result()
    """

    def on_done(self, future, **kwargs):
        future_exception = None
        try:
            future.result()
        except Exception as e:
            future_exception = e
        # If the result propogates an error, call the on_failure
        # method instead.
        if future_exception:
            self._on_failure(future, future_exception)
        else:
            self._on_success(future)

    def _on_success(self, future):
        pass

    def _on_failure(self, future, e):
        pass


class ProvideSizeSubscriber(BaseSubscriber):
    """
    A subscriber which provides the transfer size before it's queued.
    """

    def __init__(self, size):
        self.size = size

    def on_queued(self, future, **kwargs):
        # The CRT transfer client does not support providing an expected
        # size for the transfer. So only provide it if it is available.
        if hasattr(future.meta, 'provide_transfer_size'):
            future.meta.provide_transfer_size(self.size)
        else:
            LOGGER.debug(
                'Not providing transfer size. Future: %s does not offer'
                'the capability to notify the size of a transfer',
                future,
            )


class ProvideETagSubscriber(BaseSubscriber):
    """
    A subscriber which provides the object ETag before it's queued.
    """

    def __init__(self, etag):
        self.etag = etag

    def on_queued(self, future, **kwargs):
        # The CRT transfer client does not support providing an expected
        # ETag of an object. So only provide it if it is available.
        if hasattr(future.meta, 'provide_object_etag'):
            future.meta.provide_object_etag(self.etag)
        else:
            LOGGER.debug(
                'Not providing object ETag. Future: %s does not offer'
                'the capability to notify the ETag of an object',
                future,
            )


class CaseConflictCleanupSubscriber(BaseSubscriber):
    """
    A subscriber which removes object compare key from case conflict set
    when download finishes.
    """

    def __init__(self, submitted, case_conflict_key):
        self._submitted = submitted
        self._key = case_conflict_key

    def on_done(self, future, **kwargs):
        self._submitted.discard(self._key)


class DeleteSourceSubscriber(OnDoneFilteredSubscriber):
    """A subscriber which deletes the source of the transfer."""

    def _on_success(self, future):
        try:
            self._delete_source(future)
        except Exception as e:
            future.set_exception(e)

    def _delete_source(self, future):
        raise NotImplementedError('_delete_source()')


class DeleteSourceObjectSubscriber(DeleteSourceSubscriber):
    """A subscriber which deletes an object."""

    def __init__(self, client):
        self._client = client

    def _get_bucket(self, call_args):
        return call_args.bucket

    def _get_key(self, call_args):
        return call_args.key

    def _delete_source(self, future):
        call_args = future.meta.call_args
        delete_object_kwargs = {
            'Bucket': self._get_bucket(call_args),
            'Key': self._get_key(call_args),
        }
        if call_args.extra_args.get('RequestPayer'):
            delete_object_kwargs['RequestPayer'] = call_args.extra_args[
                'RequestPayer'
            ]
        self._client.delete_object(**delete_object_kwargs)


class DeleteCopySourceObjectSubscriber(DeleteSourceObjectSubscriber):
    """A subscriber which deletes the copy source."""

    def _get_bucket(self, call_args):
        return call_args.copy_source['Bucket']

    def _get_key(self, call_args):
        return call_args.copy_source['Key']


class DeleteSourceFileSubscriber(DeleteSourceSubscriber):
    """A subscriber which deletes a file."""

    def _delete_source(self, future):
        os.remove(future.meta.call_args.fileobj)


class BaseProvideContentTypeSubscriber(BaseSubscriber):
    """A subscriber that provides content type when creating s3 objects"""

    def on_queued(self, future, **kwargs):
        guessed_type = utils.guess_content_type(self._get_filename(future))
        if guessed_type is not None:
            future.meta.call_args.extra_args['ContentType'] = guessed_type

    def _get_filename(self, future):
        raise NotImplementedError('_get_filename()')


class ProvideUploadContentTypeSubscriber(BaseProvideContentTypeSubscriber):
    def _get_filename(self, future):
        return future.meta.call_args.fileobj


class ProvideLastModifiedTimeSubscriber(OnDoneFilteredSubscriber):
    """Sets utime for a downloaded file"""

    def __init__(self, last_modified_time, result_queue):
        self._last_modified_time = last_modified_time
        self._result_queue = result_queue

    def _on_success(self, future, **kwargs):
        filename = future.meta.call_args.fileobj
        try:
            last_update_tuple = self._last_modified_time.timetuple()
            mod_timestamp = time.mktime(last_update_tuple)
            utils.set_file_utime(filename, int(mod_timestamp))
        except Exception as e:
            warning_message = (
                'Successfully Downloaded %s but was unable to update the '
                'last modified time. %s' % (filename, e)
            )
            self._result_queue.put(
                utils.create_warning(filename, warning_message)
            )


class DirectoryCreatorSubscriber(BaseSubscriber):
    """Creates a directory to download if it does not exist"""

    def on_queued(self, future, **kwargs):
        d = os.path.dirname(future.meta.call_args.fileobj)
        try:
            if not os.path.exists(d):
                os.makedirs(d)
        except OSError as e:
            if not e.errno == errno.EEXIST:
                raise CreateDirectoryError(
                    "Could not create directory %s: %s" % (d, e)
                )


class CopyPropsSubscriberFactory:
    def __init__(self, client, transfer_config, cli_params):
        self._client = client
        self._transfer_config = transfer_config
        self._cli_params = cli_params

    def get_subscribers(self, fileinfo):
        copy_props = self._cli_params.get('copy_props', 'default').replace(
            '-', '_'
        )
        return getattr(self, '_get_%s_subscribers' % copy_props)(fileinfo)

    def _get_none_subscribers(self, fileinfo):
        return [
            ReplaceMetadataDirectiveSubscriber(),
            ReplaceTaggingDirectiveSubscriber(),
        ]

    def _get_metadata_directive_subscribers(self, fileinfo):
        return [
            self._create_metadata_directive_props_subscriber(fileinfo),
            ReplaceTaggingDirectiveSubscriber(),
        ]

    def _get_default_subscribers(self, fileinfo):
        return [
            self._create_metadata_directive_props_subscriber(fileinfo),
            SetTagsSubscriber(
                self._client,
                self._transfer_config,
                self._cli_params,
                source_client=fileinfo.source_client,
            ),
        ]

    def _create_metadata_directive_props_subscriber(self, fileinfo):
        subscriber_kwargs = {
            'client': fileinfo.source_client,
            'transfer_config': self._transfer_config,
            'cli_params': self._cli_params,
        }
        if not self._cli_params.get('dir_op'):
            subscriber_kwargs['head_object_response'] = (
                fileinfo.associated_response_data
            )
        return SetMetadataDirectivePropsSubscriber(**subscriber_kwargs)


class ReplaceDirectiveSubscriber(BaseSubscriber):
    _DIRECTIVE_PARAM = ''

    def on_queued(self, future, **kwargs):
        future.meta.call_args.extra_args[self._DIRECTIVE_PARAM] = 'REPLACE'


class ReplaceMetadataDirectiveSubscriber(ReplaceDirectiveSubscriber):
    _DIRECTIVE_PARAM = 'MetadataDirective'


class ReplaceTaggingDirectiveSubscriber(ReplaceDirectiveSubscriber):
    _DIRECTIVE_PARAM = 'TaggingDirective'


class SetMetadataDirectivePropsSubscriber(BaseSubscriber):
    _ALL_METADATA_DIRECTIVE_PROPERTIES = [
        'CacheControl',
        'ContentDisposition',
        'ContentEncoding',
        'ContentLanguage',
        'ContentType',
        'Expires',
        'Metadata',
    ]

    def __init__(
        self, client, transfer_config, cli_params, head_object_response=None
    ):
        self._client = client
        self._transfer_config = transfer_config
        self._cli_params = cli_params
        self._head_object_response = head_object_response

    def on_queued(self, future, **kwargs):
        should_inject_metadata = False
        if self._is_multipart_copy(future):
            should_inject_metadata = True
        elif self._has_explict_metadata_request_params(future):
            should_inject_metadata = True
            future.meta.call_args.extra_args['MetadataDirective'] = 'REPLACE'
        if should_inject_metadata:
            head_object_response = self._get_head_object_response(future)
            self._inject_metadata_props(future, head_object_response)

    def _has_explict_metadata_request_params(self, future):
        for param in future.meta.call_args.extra_args:
            if param in self._ALL_METADATA_DIRECTIVE_PROPERTIES:
                return True
        return False

    def _get_head_object_response(self, future):
        if self._head_object_response is not None:
            return self._head_object_response
        copy_source = future.meta.call_args.copy_source
        head_object_params = {
            'Bucket': copy_source['Bucket'],
            'Key': copy_source['Key'],
        }
        utils.RequestParamsMapper.map_head_object_params_with_copy_source_sse(
            head_object_params,
            self._cli_params,
        )
        return self._client.head_object(**head_object_params)

    def _inject_metadata_props(self, future, head_object_response):
        request_args = future.meta.call_args.extra_args
        for param in self._ALL_METADATA_DIRECTIVE_PROPERTIES:
            if param not in request_args and param in head_object_response:
                request_args[param] = head_object_response[param]

    def _is_multipart_copy(self, future):
        return future.meta.size >= self._transfer_config.multipart_threshold


class SetTagsSubscriber(OnDoneFilteredSubscriber):
    _MAX_TAGGING_HEADER_SIZE = 2 * 1024

    def __init__(self, client, transfer_config, cli_params, source_client):
        self._client = client
        self._transfer_config = transfer_config
        self._cli_params = cli_params
        self._source_client = source_client

    def on_queued(self, future, **kwargs):
        # Tags only need to be set if the operation is a multipart copy
        if not self._is_multipart_copy(future):
            return
        bucket, key = self._get_bucket_key_from_copy_source(future)
        tags = self._get_tags(bucket, key)
        if not tags:
            return
        header_value = self._serialize_to_header_value(tags)
        if self._fits_in_tagging_header(header_value):
            future.meta.call_args.extra_args['Tagging'] = header_value
        else:
            future.meta.user_context['TagSet'] = tags

    def _on_success(self, future):
        if 'TagSet' in future.meta.user_context:
            bucket = future.meta.call_args.bucket
            key = future.meta.call_args.key
            tag_set = future.meta.user_context['TagSet']
            try:
                self._put_object_tagging(bucket, key, tag_set)
            except Exception as e:
                self._delete_object(bucket, key)
                future.set_exception(e)

    def _put_object_tagging(self, bucket, key, tag_set):
        extra_args = {}
        utils.RequestParamsMapper.map_put_object_tagging_params(
            extra_args, self._cli_params
        )
        self._client.put_object_tagging(
            Bucket=bucket, Key=key, Tagging={'TagSet': tag_set}, **extra_args
        )

    def _delete_object(self, bucket, key):
        params = {'Bucket': bucket, 'Key': key}
        utils.RequestParamsMapper.map_delete_object_params(
            params, self._cli_params
        )
        self._client.delete_object(**params)

    def _get_bucket_key_from_copy_source(self, future):
        copy_source = future.meta.call_args.copy_source
        return copy_source['Bucket'], copy_source['Key']

    def _get_tags(self, bucket, key):
        extra_args = {}
        utils.RequestParamsMapper.map_get_object_tagging_params(
            extra_args, self._cli_params
        )
        get_tags_response = self._source_client.get_object_tagging(
            Bucket=bucket, Key=key, **extra_args
        )
        return get_tags_response['TagSet']

    def _fits_in_tagging_header(self, tagging_header):
        return (
            len(tagging_header.encode('utf-8'))
            <= self._MAX_TAGGING_HEADER_SIZE
        )

    def _serialize_to_header_value(self, tags):
        return percent_encode_sequence(
            [(tag['Key'], tag['Value']) for tag in tags]
        )

    def _is_multipart_copy(self, future):
        return future.meta.size >= self._transfer_config.multipart_threshold



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json
import re
import time
from collections import defaultdict
from datetime import datetime, timedelta

import colorama
from botocore.utils import datetime2timestamp, parse_timestamp
from dateutil import tz

from awscli.compat import get_stdout_text_writer
from awscli.customizations.commands import BasicCommand
from awscli.utils import is_a_tty


class BaseLogEventsFormatter:
    _TIMESTAMP_COLOR = colorama.Fore.GREEN
    _STREAM_NAME_COLOR = colorama.Fore.CYAN

    def __init__(self, output, colorize=True):
        self._output = output
        self._colorize = colorize
        if self._colorize:
            colorama.init(autoreset=True, strip=False)  # noqa

    def display_log_event(self, log_event):
        raise NotImplementedError('display_log_event()')

    def _color_if_configured(self, text, color):
        if self._colorize:
            return color + text + colorama.Style.RESET_ALL
        return text

    def _write_log_event(self, log_event):
        log_event = self._ensure_single_newline_ending(log_event)
        self._output.write(log_event)

    def _ensure_single_newline_ending(self, log):
        return log.rstrip() + '\n'


class ShortLogEventsFormatter(BaseLogEventsFormatter):
    def display_log_event(self, log_event):
        log_event = '%s %s' % (
            self._format_timestamp(log_event['timestamp']),
            log_event['message'],
        )
        self._write_log_event(log_event)

    def _format_timestamp(self, timestamp):
        return self._color_if_configured(
            timestamp.strftime("%Y-%m-%dT%H:%M:%S"), self._TIMESTAMP_COLOR
        )


class DetailedLogEventsFormatter(BaseLogEventsFormatter):
    def display_log_event(self, log_event):
        log_event = '%s %s %s' % (
            self._format_timestamp(log_event['timestamp']),
            self._color_if_configured(
                log_event['logStreamName'], self._STREAM_NAME_COLOR
            ),
            log_event['message'],
        )
        self._write_log_event(log_event)

    def _format_timestamp(self, timestamp):
        return self._color_if_configured(
            timestamp.isoformat(timespec='microseconds'), self._TIMESTAMP_COLOR
        )


class PrettyJSONLogEventsFormatter(BaseLogEventsFormatter):
    def display_log_event(self, log_event):
        log_event = '%s %s %s' % (
            self._format_timestamp(log_event['timestamp']),
            self._color_if_configured(
                log_event['logStreamName'], self._STREAM_NAME_COLOR
            ),
            self._format_pretty_json(log_event['message']),
        )
        self._write_log_event(log_event)

    def _format_pretty_json(self, log_message):
        try:
            loaded_json = json.loads(log_message)
            return '\n%s' % json.dumps(loaded_json, indent=4)
        except json.decoder.JSONDecodeError:
            pass
        return log_message

    def _format_timestamp(self, timestamp):
        return self._color_if_configured(
            timestamp.isoformat(), self._TIMESTAMP_COLOR
        )


class TailCommand(BasicCommand):
    NAME = 'tail'
    DESCRIPTION = (
        'Tails the logs for a CloudWatch Logs group. By default, the command '
        'returns logs from all associated CloudWatch Logs streams during the '
        'past ten minutes. Note that there is no guarantee for exact '
        'timestamp ordering of logs.'
    )
    ARG_TABLE = [
        {
            'name': 'group_name',
            'positional_arg': True,
            'help_text': 'The name of the CloudWatch Logs group.',
        },
        {
            'name': 'since',
            'default': '10m',
            'help_text': (
                'From what time to begin displaying logs. By default, logs '
                'will be displayed starting from ten minutes in the past. '
                'The value provided can be an ISO 8601 timestamp or a '
                'relative time. For relative times, provide a number and a '
                'single unit. Supported units include:\n\n'
                '<ul>'
                '<li> ``s`` - seconds </li>'
                '<li> ``m`` - minutes </li>'
                '<li> ``h`` - hours </li>'
                '<li> ``d`` - days </li>'
                '<li> ``w`` - weeks </li>'
                '</ul>'
                'For example, a value of ``5m`` would indicate to '
                'display logs starting five minutes in the past. '
                'Note that multiple units are **not** supported '
                '(i.e. ``5h30m``)'
            ),
        },
        {
            'name': 'follow',
            'action': 'store_true',
            'default': False,
            'help_text': (
                'Whether to continuously poll for new logs. By default, the '
                'command will exit once there are no more logs to display. '
                'To exit from this mode, use Control-C.'
            ),
        },
        {
            'name': 'format',
            'default': 'detailed',
            'choices': ['detailed', 'short', 'json'],
            'help_text': (
                'The format to display the logs. The following formats are '
                'supported:\n\n'
                '<ul>'
                '<li> detailed - This the default format. It prints out the '
                'timestamp with millisecond precision and timezones, the '
                'log stream name, and the log message.'
                '</li>'
                '<li> short - A shortened format. It prints out the '
                'a shortened timestamp and the log message.'
                '</li>'
                '<li> json - Pretty print any messages that are entirely JSON.'
                '</li>'
                '</ul>'
            ),
        },
        {
            'name': 'filter-pattern',
            'help_text': (
                'The filter pattern to use. See '
                '<a href="http://docs.aws.amazon.com/AmazonCloudWatch/'
                'latest/logs/FilterAndPatternSyntax.html">Filter and '
                'Pattern Syntax</a> for details. If not provided, all '
                'the events are matched'
            ),
        },
        {
            'name': 'log-stream-names',
            'nargs': '+',
            'help_text': (
                'The list of stream names to filter logs by. This parameter '
                'cannot be specified when ``--log-stream-name-prefix`` is '
                'also specified.'
            ),
        },
        {
            'name': 'log-stream-name-prefix',
            'help_text': (
                'The prefix to filter logs by. Only events from log streams '
                'with names beginning with this prefix will be returned. This '
                'parameter cannot be specified when ``log-stream-names`` is '
                'also specified.'
            ),
        },
    ]
    _FORMAT_TO_FORMATTER_CLS = {
        'detailed': DetailedLogEventsFormatter,
        'short': ShortLogEventsFormatter,
        'json': PrettyJSONLogEventsFormatter,
    }

    def _run_main(self, parsed_args, parsed_globals):
        logs_client = self._session.create_client(
            'logs',
            region_name=parsed_globals.region,
            endpoint_url=parsed_globals.endpoint_url,
            verify=parsed_globals.verify_ssl,
        )
        logs_generator = self._get_log_events_generator(
            logs_client, parsed_args.follow
        )
        log_events = logs_generator.iter_log_events(
            parsed_args.group_name,
            start=parsed_args.since,
            filter_pattern=parsed_args.filter_pattern,
            log_stream_names=parsed_args.log_stream_names,
            log_stream_name_prefix=parsed_args.log_stream_name_prefix,
        )
        self._output_log_events(parsed_args, parsed_globals, log_events)
        return 0

    def _get_log_events_generator(self, logs_client, follow):
        timestamp_utils = TimestampUtils()
        if follow:
            return FollowLogEventsGenerator(logs_client, timestamp_utils)
        else:
            return NoFollowLogEventsGenerator(logs_client, timestamp_utils)

    def _output_log_events(self, parsed_args, parsed_globals, log_events):
        output = get_stdout_text_writer()
        logs_formatter = self._FORMAT_TO_FORMATTER_CLS[parsed_args.format](
            output, colorize=self._should_use_color(parsed_globals)
        )
        for event in log_events:
            logs_formatter.display_log_event(event)

    def _should_use_color(self, parsed_globals):
        if parsed_globals.color == 'on':
            return True
        elif parsed_globals.color == 'off':
            return False
        return is_a_tty()


class TimestampUtils:
    _RELATIVE_TIMESTAMP_REGEX = re.compile(
        r"(?P<amount>\d+)(?P<unit>s|m|h|d|w)$"
    )
    _TO_SECONDS = {
        's': 1,
        'm': 60,
        'h': 3600,
        'd': 24 * 3600,
        'w': 7 * 24 * 3600,
    }

    def __init__(self, now=None):
        self._now = now
        if now is None:
            self._now = datetime.utcnow

    def to_epoch_millis(self, timestamp):
        re_match = self._RELATIVE_TIMESTAMP_REGEX.match(timestamp)
        if re_match:
            datetime_value = self._relative_timestamp_to_datetime(
                int(re_match.group('amount')), re_match.group('unit')
            )
        else:
            datetime_value = parse_timestamp(timestamp)
        return int(datetime2timestamp(datetime_value) * 1000)

    def to_datetime(self, millis_timestamp):
        return datetime.fromtimestamp(millis_timestamp / 1000.0, tz.tzutc())

    def _relative_timestamp_to_datetime(self, amount, unit):
        multiplier = self._TO_SECONDS[unit]
        return self._now() + timedelta(seconds=amount * multiplier * -1)


class BaseLogEventsGenerator:
    def __init__(self, client, timestamp_utils):
        self._client = client
        self._timestamp_utils = timestamp_utils

    def iter_log_events(
        self,
        group_name,
        start=None,
        filter_pattern=None,
        log_stream_names=None,
        log_stream_name_prefix=None,
    ):
        filter_logs_events_kwargs = self._get_filter_logs_events_kwargs(
            group_name,
            start,
            filter_pattern,
            log_stream_names,
            log_stream_name_prefix,
        )
        log_events = self._filter_log_events(filter_logs_events_kwargs)
        for log_event in log_events:
            self._convert_event_timestamps(log_event)
            yield log_event

    def _filter_log_events(self, filter_logs_events_kwargs):
        raise NotImplementedError('_filter_log_events()')

    def _get_filter_logs_events_kwargs(
        self,
        group_name,
        start,
        filter_pattern,
        log_stream_names,
        log_stream_name_prefix,
    ):
        kwargs = {'logGroupName': group_name, 'interleaved': True}
        if start is not None:
            kwargs['startTime'] = self._timestamp_utils.to_epoch_millis(start)
        if filter_pattern is not None:
            kwargs['filterPattern'] = filter_pattern
        if log_stream_names is not None:
            kwargs['logStreamNames'] = log_stream_names
        if log_stream_name_prefix is not None:
            kwargs['logStreamNamePrefix'] = log_stream_name_prefix
        return kwargs

    def _convert_event_timestamps(self, event):
        event['ingestionTime'] = self._timestamp_utils.to_datetime(
            event['ingestionTime']
        )
        event['timestamp'] = self._timestamp_utils.to_datetime(
            event['timestamp']
        )


class NoFollowLogEventsGenerator(BaseLogEventsGenerator):
    def _filter_log_events(self, filter_logs_events_kwargs):
        paginator = self._client.get_paginator('filter_log_events')
        for page in paginator.paginate(**filter_logs_events_kwargs):
            for log_event in page['events']:
                yield log_event


class FollowLogEventsGenerator(BaseLogEventsGenerator):
    _TIME_TO_SLEEP = 5

    def __init__(self, client, timestamp_utils, sleep=None):
        super(FollowLogEventsGenerator, self).__init__(client, timestamp_utils)
        self._sleep = sleep
        if sleep is None:
            self._sleep = time.sleep

    def _filter_log_events(self, filter_logs_events_kwargs):
        try:
            for event in self._do_filter_log_events(filter_logs_events_kwargs):
                yield event
        except KeyboardInterrupt:
            # The only way to exit from the --follow is to Ctrl-C. So
            # we should exit the iterator rather than having the
            # KeyboardInterrupt propogate to the rest of the command.
            return

    def _get_latest_events_and_timestamp(self, event_ids_per_timestamp):
        if event_ids_per_timestamp:
            # Keep only ids of the events with the newest timestamp
            newest_timestamp = max(event_ids_per_timestamp.keys())
            event_ids_per_timestamp = defaultdict(
                set,
                {newest_timestamp: event_ids_per_timestamp[newest_timestamp]},
            )
        return event_ids_per_timestamp

    def _reset_filter_log_events_params(
        self, fle_kwargs, event_ids_per_timestamp
    ):
        # Remove nextToken and update startTime for the next request
        # with the timestamp of the newest event
        if event_ids_per_timestamp:
            fle_kwargs['startTime'] = max(event_ids_per_timestamp.keys())
        fle_kwargs.pop('nextToken', None)

    def _do_filter_log_events(self, filter_logs_events_kwargs):
        event_ids_per_timestamp = defaultdict(set)
        while True:
            response = self._client.filter_log_events(
                **filter_logs_events_kwargs
            )
            for event in response['events']:
                # For the case where we've hit the last page, we will be
                # reusing the newest timestamp of the received events to keep polling.
                # This means it is possible that duplicate log events with same timestamp
                # are returned back which we do not want to yield again.
                # We only want to yield log events that we have not seen.
                if (
                    event['eventId']
                    not in event_ids_per_timestamp[event['timestamp']]
                ):
                    event_ids_per_timestamp[event['timestamp']].add(
                        event['eventId']
                    )
                    yield event
            event_ids_per_timestamp = self._get_latest_events_and_timestamp(
                event_ids_per_timestamp
            )
            if 'nextToken' in response:
                filter_logs_events_kwargs['nextToken'] = response['nextToken']
            else:
                self._reset_filter_log_events_params(
                    filter_logs_events_kwargs, event_ids_per_timestamp
                )
                self._sleep(self._TIME_TO_SLEEP)



# Copyright 2012-2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import re

import ruamel.yaml
from botocore.compat import OrderedDict, json
from ruamel.yaml.resolver import ScalarNode, SequenceNode

from awscli.utils import SafeLineBreakEmitter, dump_yaml_to_str


def intrinsics_multi_constructor(loader, tag_prefix, node):
    """
    YAML constructor to parse CloudFormation intrinsics.
    This will return a dictionary with key being the intrinsic name
    """

    # Get the actual tag name excluding the first exclamation
    tag = node.tag[1:]

    # Some intrinsic functions doesn't support prefix "Fn::"
    prefix = "Fn::"
    if tag in ["Ref", "Condition"]:
        prefix = ""

    cfntag = prefix + tag

    if tag == "GetAtt" and isinstance(node.value, str):
        # ShortHand notation for !GetAtt accepts Resource.Attribute format
        # while the standard notation is to use an array
        # [Resource, Attribute]. Convert shorthand to standard format
        value = node.value.split(".", 1)

    elif isinstance(node, ScalarNode):
        # Value of this node is scalar
        value = loader.construct_scalar(node)

    elif isinstance(node, SequenceNode):
        # Value of this node is an array (Ex: [1,2])
        value = loader.construct_sequence(node)

    else:
        # Value of this node is an mapping (ex: {foo: bar})
        value = loader.construct_mapping(node)

    return {cfntag: value}


def _dict_representer(dumper, data):
    return dumper.represent_dict(data.items())


def _add_yaml_1_1_boolean_resolvers(resolver_cls):
    # CloudFormation treats unquoted values that are YAML 1.1 native
    # booleans as booleans, rather than strings. In YAML 1.2, the only
    # boolean values are "true" and "false" so values such as "yes" and "no"
    # when loaded as strings are not quoted when dumped. This logic ensures
    # that we dump these values with quotes so that CloudFormation treats
    # these values as strings and not booleans.
    boolean_regex = re.compile(
        '^(?:yes|Yes|YES|no|No|NO'
        '|true|True|TRUE|false|False|FALSE'
        '|on|On|ON|off|Off|OFF)$'
    )
    boolean_first_chars = list('yYnNtTfFoO')
    resolver_cls.add_implicit_resolver(
        'tag:yaml.org,2002:bool', boolean_regex, boolean_first_chars
    )


def yaml_dump(dict_to_dump):
    """
    Dumps the dictionary as a YAML document
    :param dict_to_dump:
    :return:
    """

    yaml = ruamel.yaml.YAML(typ="safe", pure=True)
    yaml.default_flow_style = False
    yaml.Emitter = SafeLineBreakEmitter
    yaml.Representer = FlattenAliasRepresenter
    _add_yaml_1_1_boolean_resolvers(yaml.Resolver)
    yaml.Representer.add_representer(OrderedDict, _dict_representer)

    return dump_yaml_to_str(yaml, dict_to_dump)


def _dict_constructor(loader, node):
    # Necessary in order to make yaml merge tags work
    loader.flatten_mapping(node)
    return OrderedDict(loader.construct_pairs(node))


def yaml_parse(yamlstr):
    """Parse a yaml string"""
    try:
        # PyYAML doesn't support json as well as it should, so if the input
        # is actually just json it is better to parse it with the standard
        # json parser.
        return json.loads(yamlstr, object_pairs_hook=OrderedDict)
    except ValueError:
        yaml = ruamel.yaml.YAML(typ="safe", pure=True)
        yaml.Constructor.add_constructor(
            ruamel.yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
            _dict_constructor,
        )
        yaml.Constructor.add_multi_constructor(
            "!", intrinsics_multi_constructor
        )
        _add_yaml_1_1_boolean_resolvers(yaml.Resolver)

        return yaml.load(yamlstr)


class FlattenAliasRepresenter(ruamel.yaml.representer.SafeRepresenter):
    def ignore_aliases(self, data):
        return True



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

"""Display strategy implementations for ECS Express Gateway Service monitoring."""

import asyncio
import time

from botocore.exceptions import ClientError
from colorama import Style

from awscli.customizations.ecs.exceptions import MonitoringError
from awscli.customizations.ecs.expressgateway.stream_display import (
    StreamDisplay,
)
from awscli.customizations.utils import uni_print


class DisplayStrategy:
    """Base class for display strategies.

    Each strategy controls its own execution model, timing, and output format.
    """

    def execute_monitoring(self, collector, start_time, timeout_minutes):
        """Execute the monitoring loop.

        Args:
            collector: ServiceViewCollector instance for data fetching
            start_time: Start timestamp for timeout calculation
            timeout_minutes: Maximum monitoring duration in minutes
        """
        raise NotImplementedError


class InteractiveDisplayStrategy(DisplayStrategy):
    """Interactive display strategy with async spinner and keyboard navigation.

    Uses dual async tasks:
    - Data task: Polls ECS APIs every 5 seconds
    - Spinner task: Updates display every 100ms with rotating spinner
    """

    def __init__(self, display, use_color):
        """Initialize the interactive display strategy.

        Args:
            display: Display instance from prompt_toolkit_display module
                providing the interactive terminal interface
            use_color: Whether to use colored output
        """
        self.display = display
        self.use_color = use_color

    def execute_monitoring(self, collector, start_time, timeout_minutes):
        """Execute async monitoring with spinner and keyboard controls."""
        try:
            final_output, timed_out = asyncio.run(
                self._execute_async(collector, start_time, timeout_minutes)
            )
            if timed_out:
                uni_print(final_output + "\nMonitoring timed out!\n")
            else:
                uni_print(final_output + "\nMonitoring Complete!\n")
        finally:
            uni_print(Style.RESET_ALL)

    async def _execute_async(self, collector, start_time, timeout_minutes):
        """Async execution with dual tasks for data and spinner."""
        spinner_chars = "⠋⠙⠹⠸⠼⠴⠦⠧⠇⠏"
        spinner_index = 0
        current_output = "Waiting for initial data"
        timed_out = False

        async def update_data():
            nonlocal current_output, timed_out
            while True:
                current_time = time.time()
                if current_time - start_time > timeout_minutes * 60:
                    timed_out = True
                    # Only exit if app is running to avoid "Application is not running" error
                    if self.display.app.is_running:
                        self.display.app.exit()
                    break

                try:
                    loop = asyncio.get_event_loop()
                    new_output = await loop.run_in_executor(
                        None, collector.get_current_view, "{SPINNER}"
                    )
                    current_output = new_output
                except ClientError as e:
                    if (
                        e.response.get('Error', {}).get('Code')
                        == 'InvalidParameterException'
                    ):
                        error_message = e.response.get('Error', {}).get(
                            'Message', ''
                        )
                        if (
                            "Cannot call DescribeServiceRevisions for a service that is INACTIVE"
                            in error_message
                        ):
                            current_output = "Service is inactive"
                        else:
                            raise
                    else:
                        raise

                await asyncio.sleep(5.0)

        async def update_spinner():
            nonlocal spinner_index
            while True:
                spinner_char = spinner_chars[spinner_index]
                display_output = current_output.replace(
                    "{SPINNER}", spinner_char
                )
                status_text = f"Getting updates... {spinner_char} | up/down to scroll, q to quit"
                self.display.display(display_output, status_text)
                spinner_index = (spinner_index + 1) % len(spinner_chars)
                await asyncio.sleep(0.1)

        data_task = asyncio.create_task(update_data())
        spinner_task = asyncio.create_task(update_spinner())
        display_task = None

        try:
            display_task = asyncio.create_task(self.display.run())

            done, pending = await asyncio.wait(
                [display_task, data_task], return_when=asyncio.FIRST_COMPLETED
            )

            if data_task in done:
                # Retrieve and re-raise any exception from the task.
                # asyncio.wait() doesn't retrieve exceptions itself.
                exc = data_task.exception()
                if exc:
                    raise exc

            # Cancel pending tasks
            for task in pending:
                task.cancel()
                # Await cancelled task to ensure proper cleanup and prevent
                # warnings about unawaited tasks
                try:
                    await task
                except asyncio.CancelledError:
                    pass

        finally:
            # Ensure display app is properly shut down
            # Only exit if app is running to avoid "Application is not running" error
            if self.display.app.is_running:
                self.display.app.exit()
            spinner_task.cancel()
            if display_task is not None and not display_task.done():
                display_task.cancel()
                # Await cancelled task to ensure proper cleanup and prevent
                # warnings about unawaited tasks
                try:
                    await display_task
                except asyncio.CancelledError:
                    pass

        return current_output.replace("{SPINNER}", ""), timed_out


class TextOnlyDisplayStrategy(DisplayStrategy):
    """Text-only display strategy with diff detection and timestamped output.

    Uses synchronous polling loop with change detection to output only
    individual resource changes with timestamps.
    """

    def __init__(self, use_color):
        self.stream_display = StreamDisplay(use_color)

    def execute_monitoring(self, collector, start_time, timeout_minutes):
        """Execute synchronous monitoring with text output."""
        self.stream_display.show_startup_message()

        try:
            while True:
                current_time = time.time()
                if current_time - start_time > timeout_minutes * 60:
                    self.stream_display.show_timeout_message()
                    break

                try:
                    collector.get_current_view("")

                    # Extract cached result for diff detection
                    managed_resources, info = collector.cached_monitor_result

                    self.stream_display.show_monitoring_data(
                        managed_resources, info
                    )

                except ClientError as e:
                    if (
                        e.response.get('Error', {}).get('Code')
                        == 'InvalidParameterException'
                    ):
                        error_message = e.response.get('Error', {}).get(
                            'Message', ''
                        )
                        if (
                            "Cannot call DescribeServiceRevisions for a service that is INACTIVE"
                            in error_message
                        ):
                            self.stream_display.show_service_inactive_message()
                            break
                        else:
                            raise
                    else:
                        raise

                time.sleep(5.0)

        except KeyboardInterrupt:
            self.stream_display.show_user_stop_message()
        except MonitoringError as e:
            self.stream_display.show_error_message(e)
        finally:
            self.stream_display.show_completion_message()
            uni_print(Style.RESET_ALL)



# Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.customizations.ec2instanceconnect.opentunnel import (
    OpenTunnelCommand,
)
from awscli.customizations.ec2instanceconnect.ssh import (
    SshCommand,
)


def register_ec2_instance_connect_commands(event_handlers):
    """
    The entry point for EC2 Instance Connect high level commands.
    """
    event_handlers.register(
        "building-command-table.ec2-instance-connect", inject_commands
    )


def inject_commands(command_table, session, **kwargs):
    """
    Called when the EC2 Instance Connect command table is being built.
    Used to inject new high level commands into the command list.
    """
    command_table["open-tunnel"] = OpenTunnelCommand(session)
    command_table["ssh"] = SshCommand(session)



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging

from awscli.customizations.exceptions import ParamValidationError

LOG = logging.getLogger(__name__)

VALID_SYNC_TYPES = [
    'file_at_src_and_dest',
    'file_not_at_dest',
    'file_not_at_src',
]


class BaseSync:
    """Base sync strategy

    To create a new sync strategy, subclass from this class.
    """

    # This is the argument that will be added to the ``SyncCommand`` arg table.
    # This argument will represent the sync strategy when the arguments for
    # the sync command are parsed.  ``ARGUMENT`` follows the same format as
    # a member of ``ARG_TABLE`` in ``BasicCommand`` class as specified in
    # ``awscli/customizations/commands.py``.
    #
    # For example, if I wanted to perform the sync strategy whenever I type
    # ``--my-sync-strategy``, I would say:
    #
    # ARGUMENT =
    #     {'name': 'my-sync-strategy', 'action': 'store-true',
    #      'help_text': 'Performs my sync strategy'}
    #
    # Typically, the argument's ``action`` should ``store_true`` to
    # minimize amount of extra code in making a custom sync strategy.
    ARGUMENT = None

    # At this point all that need to be done is implement
    # ``determine_should_sync`` method (see method for more information).

    def __init__(self, sync_type='file_at_src_and_dest'):
        """
        :type sync_type: string
        :param sync_type: This determines where the sync strategy will be
            used. There are three strings to choose from:

            'file_at_src_and_dest': apply sync strategy on a file that
            exists both at the source and the destination.

            'file_not_at_dest': apply sync strategy on a file that
            exists at the source but not the destination.

            'file_not_at_src': apply sync strategy on a file that
            exists at the destination but not the source.
        """
        self._check_sync_type(sync_type)
        self._sync_type = sync_type

    def _check_sync_type(self, sync_type):
        if sync_type not in VALID_SYNC_TYPES:
            raise ParamValidationError(
                "Unknown sync_type: %s.\n"
                "Valid options are %s." % (sync_type, VALID_SYNC_TYPES)
            )

    @property
    def sync_type(self):
        return self._sync_type

    def register_strategy(self, session):
        """Registers the sync strategy class to the given session."""

        session.register('building-arg-table.s3_sync', self.add_sync_argument)
        session.register('choosing-s3-sync-strategy', self.use_sync_strategy)

    def determine_should_sync(self, src_file, dest_file):
        """Subclasses should implement this method.

        This function takes two ``FileStat`` objects (one from the source and
        one from the destination).  Then makes a decision on whether a given
        operation (e.g. a upload, copy, download) should be allowed
        to take place.

        The function currently raises a ``NotImplementedError``.  So this
        method must be overwritten when this class is subclassed.  Note
        that this method must return a Boolean as documented below.

        :type src_file: ``FileStat`` object
        :param src_file: A representation of the operation that is to be
            performed on a specific file existing in the source.  Note if
            the file does not exist at the source, ``src_file`` is None.

        :type dest_file: ``FileStat`` object
        :param dest_file: A representation of the operation that is to be
            performed on a specific file existing in the destination. Note if
            the file does not exist at the destination, ``dest_file`` is None.

        :rtype: Boolean
        :return: True if an operation based on the ``FileStat`` should be
            allowed to occur.
            False if if an operation based on the ``FileStat`` should not be
            allowed to occur. Note the operation being referred to depends on
            the ``sync_type`` of the sync strategy:

            'file_at_src_and_dest': refers to ``src_file``

            'file_not_at_dest': refers to ``src_file``

            'file_not_at_src': refers to ``dest_file``
        """

        raise NotImplementedError("determine_should_sync")

    @property
    def arg_name(self):
        # Retrieves the ``name`` of the sync strategy's ``ARGUMENT``.
        name = None
        if self.ARGUMENT is not None:
            name = self.ARGUMENT.get('name', None)
        return name

    @property
    def arg_dest(self):
        # Retrieves the ``dest`` of the sync strategy's ``ARGUMENT``.
        dest = None
        if self.ARGUMENT is not None:
            dest = self.ARGUMENT.get('dest', None)
        return dest

    def add_sync_argument(self, arg_table, **kwargs):
        # This function adds sync strategy's argument to the ``SyncCommand``
        # argument table.
        if self.ARGUMENT is not None:
            arg_table.append(self.ARGUMENT)

    def use_sync_strategy(self, params, **kwargs):
        # This function determines which sync strategy the ``SyncCommand`` will
        # use. The sync strategy object must be returned by this method
        # if it is to be chosen as the sync strategy to use.
        #
        # ``params`` is a dictionary that specifies all of the arguments
        # the sync command is able to process as well as their values.
        #
        # Since ``ARGUMENT`` was added to the ``SyncCommand`` arg table,
        # the argument will be present in ``params``.
        #
        # If the argument was included in the actual ``aws s3 sync`` command
        # its value will show up as ``True`` in ``params`` otherwise its value
        # will be ``False`` in ``params`` assuming the argument's ``action``
        # is ``store_true``.
        #
        # Note: If the ``action`` of ``ARGUMENT`` was not set to
        # ``store_true``, this method will need to be overwritten.
        #
        name_in_params = None
        # Check if a ``dest`` was specified in ``ARGUMENT`` as if it is
        # specified, the boolean value will be located at the argument's
        # ``dest`` value in the ``params`` dictionary.
        if self.arg_dest is not None:
            name_in_params = self.arg_dest
        # Then check ``name`` of ``ARGUMENT``, the boolean value will be
        # located at the argument's ``name`` value in the ``params``
        # dictionary.
        elif self.arg_name is not None:
            # ``name`` has all ``-`` replaced with ``_`` in ``params``.
            name_in_params = self.arg_name.replace('-', '_')
        if name_in_params is not None:
            if params.get(name_in_params):
                # Return the sync strategy object to be used for syncing.
                return self
        return None

    def total_seconds(self, td):
        """
        timedelta's time_seconds() function for python 2.6 users

        :param td: The difference between two datetime objects.
        """
        return (
            td.microseconds + (td.seconds + td.days * 24 * 3600) * 10**6
        ) / 10**6

    def compare_size(self, src_file, dest_file):
        """
        :returns: True if the sizes are the same.
            False otherwise.
        """
        return src_file.size == dest_file.size

    def compare_time(self, src_file, dest_file):
        """
        :returns: True if the file does not need updating based on time of
            last modification and type of operation.
            False if the file does need updating based on the time of
            last modification and type of operation.
        """
        src_time = src_file.last_update
        dest_time = dest_file.last_update
        delta = dest_time - src_time
        cmd = src_file.operation_name
        if cmd == "upload" or cmd == "copy":
            if self.total_seconds(delta) >= 0:
                # Destination is newer than source.
                return True
            else:
                # Destination is older than source, so
                # we have a more recently updated file
                # at the source location.
                return False
        elif cmd == "download":
            if self.total_seconds(delta) <= 0:
                return True
            else:
                # delta is positive, so the destination
                # is newer than the source.
                return False


class SizeAndLastModifiedSync(BaseSync):
    def determine_should_sync(self, src_file, dest_file):
        same_size = self.compare_size(src_file, dest_file)
        same_last_modified_time = self.compare_time(src_file, dest_file)
        should_sync = (not same_size) or (not same_last_modified_time)
        if should_sync:
            LOG.debug(
                "syncing: %s -> %s, size: %s -> %s, modified time: %s -> %s",
                src_file.src,
                src_file.dest,
                src_file.size,
                dest_file.size,
                src_file.last_update,
                dest_file.last_update,
            )
        return should_sync


class NeverSync(BaseSync):
    def __init__(self, sync_type='file_not_at_src'):
        super(NeverSync, self).__init__(sync_type)

    def determine_should_sync(self, src_file, dest_file):
        return False


class MissingFileSync(BaseSync):
    def __init__(self, sync_type='file_not_at_dest'):
        super(MissingFileSync, self).__init__(sync_type)

    def determine_should_sync(self, src_file, dest_file):
        LOG.debug(
            "syncing: %s -> %s, file does not exist at destination",
            src_file.src,
            src_file.dest,
        )
        return True


class AlwaysSync(BaseSync):
    def __init__(self, sync_type='file_at_src_and_dest'):
        super(AlwaysSync, self).__init__(sync_type)

    def determine_should_sync(self, src_file, dest_file):
        LOG.debug(f"syncing: {src_file.src} -> {src_file.dest}")
        return True



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from prompt_toolkit import Application
from prompt_toolkit.application import get_app
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.key_binding.key_bindings import KeyBindings
from prompt_toolkit.layout import Float, FloatContainer, Layout
from prompt_toolkit.layout.containers import ScrollOffsets, Window
from prompt_toolkit.layout.controls import UIContent, UIControl
from prompt_toolkit.layout.dimension import Dimension
from prompt_toolkit.layout.margins import ScrollbarMargin
from prompt_toolkit.layout.screen import Point
from prompt_toolkit.utils import get_cwidth


def select_menu(items, display_format=None, max_height=10):
    """Presents a list of options and allows the user to select one.

    This presents a static list of options and prompts the user to select one.
    This is similar to a completion menu but is different in that it does not
    allow a user to type and the returned value is always a member of the list.

    :type items: list
    :param list: The list of items to be selected from. If this list contains
    elements that are not strings the display_format option must be specified.

    :type display_format: Callable[[Any], str]
    :param display_format: A callable that takes a single element from the
    items list as input and returns a string used to represent the item in the
    menu.

    :type max_height: int
    :param max_height: The max number of items to show in the list at a time.

    :returns: The selected element from the items list.
    """
    app_bindings = KeyBindings()

    @app_bindings.add('c-c')
    def exit_app(event):
        event.app.exit(exception=KeyboardInterrupt, style='class:aborting')

    min_height = min(max_height, len(items))
    menu_window = Window(
        SelectionMenuControl(items, display_format=display_format),
        always_hide_cursor=False,
        height=Dimension(min=min_height, max=min_height),
        scroll_offsets=ScrollOffsets(),
        right_margins=[ScrollbarMargin()],
    )

    # Using a FloatContainer was the only way I was able to succesfully
    # limit the height and width of the window.
    content = FloatContainer(
        Window(height=Dimension(min=min_height, max=min_height)),
        [Float(menu_window, top=0, left=0)],
    )
    app = Application(
        layout=Layout(content),
        key_bindings=app_bindings,
        erase_when_done=True,
    )
    return app.run()


def _trim_text(text, max_width):
    """
    Trim the text to `max_width`, append dots when the text is too long.
    Returns (text, width) tuple.
    """
    width = get_cwidth(text)

    # When the text is too wide, trim it.
    if width > max_width:
        # When there are no double width characters, just use slice operation.
        if len(text) == width:
            trimmed_text = (text[: max(1, max_width - 3)] + '...')[:max_width]
            return trimmed_text, len(trimmed_text)

        # Otherwise, loop until we have the desired width. (Rather
        # inefficient, but ok for now.)
        else:
            trimmed_text = ''
            for c in text:
                if get_cwidth(trimmed_text + c) <= max_width - 3:
                    trimmed_text += c
            trimmed_text += '...'

            return (trimmed_text, get_cwidth(trimmed_text))
    else:
        return text, width


class SelectionMenuControl(UIControl):
    MIN_WIDTH = 7
    # The char width overhead of formatting the text into the menu

    def __init__(self, items, display_format=None, cursor='>'):
        self._items = items
        self._selection = 0
        self._cursor = cursor
        self._display_format = display_format
        self._format_overhead = 3 + len(cursor)

    def _get_items(self):
        if callable(self._items):
            self._items = self._items()
        return self._items

    def is_focusable(self):
        return True

    def preferred_width(self, max_width):
        items = self._get_items()
        if self._display_format:
            items = (self._display_format(i) for i in items)
        max_item_width = max(get_cwidth(i) for i in items)
        max_item_width += self._format_overhead
        if max_item_width < self.MIN_WIDTH:
            max_item_width = self.MIN_WIDTH
        return min(max_width, max_item_width)

    def preferred_height(self, width, max_height, wrap_lines, get_line_prefix):
        return min(max_height, len(self._get_items()))

    def _menu_item_fragment(self, item, is_selected, menu_width):
        if is_selected:
            cursor = self._cursor
            style_str = 'class:completion-menu.completion.current'
        else:
            cursor = ' ' * len(self._cursor)
            style_str = 'class:completion-menu.completion'

        if self._display_format:
            item = self._display_format(item)

        text, tw = _trim_text(item, menu_width - self._format_overhead)
        padding = ' ' * (menu_width - self._format_overhead - tw)
        return [(style_str, '%s %s%s  ' % (cursor, text, padding))]

    def create_content(self, width, height):
        def get_line(i):
            item = self._get_items()[i]
            is_selected = i == self._selection
            return self._menu_item_fragment(item, is_selected, width)

        return UIContent(
            get_line=get_line,
            cursor_position=Point(x=0, y=self._selection or 0),
            line_count=len(self._get_items()),
        )

    def _move_cursor(self, delta):
        self._selection += delta

        num_items = len(self._get_items())
        if self._selection >= num_items:
            self._selection = 0
        elif self._selection < 0:
            self._selection = num_items - 1

    def get_key_bindings(self):
        kb = KeyBindings()

        @kb.add('up')
        def move_up(event):
            self._move_cursor(-1)

        @kb.add('down')
        def move_down(event):
            self._move_cursor(1)

        @kb.add('enter')
        def app_result(event):
            result = self._get_items()[self._selection]
            event.app.exit(result=result)

        return kb


class CollapsableSelectionMenuControl(SelectionMenuControl):
    """Menu that collapses to text with selection when loses focus"""

    def __init__(
        self,
        items,
        display_format=None,
        cursor='>',
        selection_capture_buffer=None,
        on_toggle=None,
    ):
        super().__init__(items, display_format=display_format, cursor=cursor)
        if not selection_capture_buffer:
            selection_capture_buffer = Buffer()
        self.buffer = selection_capture_buffer
        self._has_ever_entered_select_menu = False
        self.on_toggle = on_toggle

    def create_content(self, width, height):
        if get_app().layout.has_focus(self):
            self._has_ever_entered_select_menu = True
            return super().create_content(width, height)
        else:

            def get_line(i):
                content = ''
                if self._has_ever_entered_select_menu:
                    content = self._get_items()[self._selection]
                return [('', content)]

            return UIContent(get_line=get_line, line_count=1)

    def preferred_height(self, width, max_height, wrap_lines, get_line_prefix):
        if get_app().layout.has_focus(self):
            return super().preferred_height(
                width, max_height, wrap_lines, get_line_prefix
            )
        else:
            return 1

    def _get_items(self):
        items = super()._get_items()
        # Initialize buffer selection text if it had not been set previously
        # (e.g. it was the first time items were retrieved)
        if items is None:
            return ['']
        if not self.buffer.text:
            self.buffer.text = items[self._selection]
            if callable(self.on_toggle):
                self.on_toggle(self.buffer.text)
        return items

    def _move_cursor(self, delta):
        super()._move_cursor(delta)
        self.buffer.text = self._get_items()[self._selection]
        if callable(self.on_toggle):
            self.on_toggle(self.buffer.text)

    def get_key_bindings(self):
        kb = KeyBindings()

        @kb.add('up')
        def move_up(event):
            self._move_cursor(-1)

        @kb.add('down')
        def move_down(event):
            self._move_cursor(1)

        return kb



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging
import os

from botocore.compat import OrderedDict

from awscli.compat import is_windows
from awscli.customizations.commands import BasicCommand
from awscli.customizations.eks.exceptions import EKSClusterError
from awscli.customizations.eks.kubeconfig import (
    Kubeconfig,
    KubeconfigAppender,
    KubeconfigError,
    KubeconfigLoader,
    KubeconfigValidator,
    KubeconfigWriter,
)
from awscli.customizations.eks.ordered_yaml import ordered_yaml_dump
from awscli.customizations.utils import uni_print

LOG = logging.getLogger(__name__)

DEFAULT_PATH = os.path.expanduser("~/.kube/config")

# At the time EKS no longer supports Kubernetes v1.21 (probably ~Dec 2023),
# this can be safely changed to default to writing "v1"
API_VERSION = "client.authentication.k8s.io/v1beta1"


class UpdateKubeconfigCommand(BasicCommand):
    NAME = 'update-kubeconfig'

    DESCRIPTION = BasicCommand.FROM_FILE(
        'eks', 'update-kubeconfig', '_description.rst'
    )

    ARG_TABLE = [
        {
            'name': 'name',
            'dest': 'cluster_name',
            'help_text': (
                "The name of the cluster for which "
                "to create a kubeconfig entry. "
                "This cluster must exist in your account and in the "
                "specified or configured default Region "
                "for your AWS CLI installation."
            ),
            'required': True,
        },
        {
            'name': 'kubeconfig',
            'help_text': (
                "Optionally specify a kubeconfig file to append "
                "with your configuration. "
                "By default, the configuration is written to the "
                "first file path in the KUBECONFIG "
                "environment variable (if it is set) "
                "or the default kubeconfig path (.kube/config) "
                "in your home directory."
            ),
            'required': False,
        },
        {
            'name': 'role-arn',
            'help_text': (
                "To assume a role for cluster authentication, "
                "specify an IAM role ARN with this option. "
                "For example, if you created a cluster "
                "while assuming an IAM role, "
                "then you must also assume that role to "
                "connect to the cluster the first time."
            ),
            'required': False,
        },
        {
            'name': 'proxy-url',
            'help_text': (
                "Optionally specify a proxy url to route "
                "traffic via when connecting to a cluster."
            ),
            'required': False,
        },
        {
            'name': 'dry-run',
            'action': 'store_true',
            'default': False,
            'help_text': (
                "Print the merged kubeconfig to stdout instead of "
                "writing it to the specified file."
            ),
            'required': False,
        },
        {
            'name': 'verbose',
            'action': 'store_true',
            'default': False,
            'help_text': (
                "Print more detailed output "
                "when writing to the kubeconfig file, "
                "including the appended entries."
            ),
        },
        {
            'name': 'alias',
            'help_text': (
                "Alias for the cluster context name. "
                "Defaults to match cluster ARN."
            ),
            'required': False,
        },
        {
            'name': 'user-alias',
            'help_text': (
                "Alias for the generated user name. "
                "Defaults to match cluster ARN."
            ),
            'required': False,
        },
        {
            'name': 'assume-role-arn',
            'help_text': (
                'To assume a role for retrieving cluster information, '
                'specify an IAM role ARN with this option. '
                'Use this for cross-account access to get cluster details '
                'from the account where the cluster resides.'
            ),
            'required': False,
        },
    ]

    def _display_entries(self, entries):
        """
        Display entries in yaml format

        :param entries: a list of OrderedDicts to be printed
        :type entries: list
        """
        uni_print("Entries:\n\n")
        for entry in entries:
            uni_print(ordered_yaml_dump(entry))
            uni_print("\n")

    def _run_main(self, parsed_args, parsed_globals):
        client = EKSClient(
            self._session,
            parsed_args=parsed_args,
            parsed_globals=parsed_globals,
        )
        new_cluster_dict = client.get_cluster_entry()
        new_user_dict = client.get_user_entry(
            user_alias=parsed_args.user_alias
        )

        config_selector = KubeconfigSelector(
            os.environ.get("KUBECONFIG", ""), parsed_args.kubeconfig
        )
        config = config_selector.choose_kubeconfig(new_cluster_dict["name"])
        updating_existing = config.has_cluster(new_cluster_dict["name"])
        appender = KubeconfigAppender()
        new_context_dict = appender.insert_cluster_user_pair(
            config, new_cluster_dict, new_user_dict, parsed_args.alias
        )

        if parsed_args.dry_run:
            uni_print(config.dump_content())
        else:
            writer = KubeconfigWriter()
            writer.write_kubeconfig(config)

            if updating_existing:
                uni_print(
                    f"Updated context {new_context_dict['name']} in {config.path}\n"
                )
            else:
                uni_print(
                    f"Added new context {new_context_dict['name']} to {config.path}\n"
                )

            if parsed_args.verbose:
                self._display_entries(
                    [new_context_dict, new_user_dict, new_cluster_dict]
                )
        return 0


class KubeconfigSelector:
    def __init__(self, env_variable, path_in, validator=None, loader=None):
        """
        Parse KUBECONFIG into a list of absolute paths.
        Also replace the empty list with DEFAULT_PATH

        :param env_variable: KUBECONFIG as a long string
        :type env_variable: string

        :param path_in: The path passed in through the CLI
        :type path_in: string or None
        """
        if validator is None:
            validator = KubeconfigValidator()
        self._validator = validator

        if loader is None:
            loader = KubeconfigLoader(validator)
        self._loader = loader

        if path_in is not None:
            # Override environment variable
            self._paths = [self._expand_path(path_in)]
        else:
            # Get the list of paths from the environment variable
            if env_variable == "":
                env_variable = DEFAULT_PATH
            self._paths = [
                self._expand_path(element)
                for element in env_variable.split(os.pathsep)
                if len(element.strip()) > 0
            ]
            if len(self._paths) == 0:
                self._paths = [DEFAULT_PATH]

    def choose_kubeconfig(self, cluster_name):
        """
        Choose which kubeconfig file to read from.
        If name is already an entry in one of the $KUBECONFIG files,
        choose that one.
        Otherwise choose the first file.

        :param cluster_name: The name of the cluster which is going to be added
        :type cluster_name: String

        :return: a chosen Kubeconfig based on above rules
        :rtype: Kubeconfig
        """
        # Search for an existing entry to update
        for candidate_path in self._paths:
            try:
                loaded_config = self._loader.load_kubeconfig(candidate_path)

                if loaded_config.has_cluster(cluster_name):
                    LOG.debug("Found entry to update at %s", candidate_path)
                    return loaded_config
            except KubeconfigError as e:
                LOG.warning("Passing %s:%s", candidate_path, e)

        # No entry was found, use the first file in KUBECONFIG
        #
        # Note: This could raise KubeconfigErrors if paths[0] is corrupted
        return self._loader.load_kubeconfig(self._paths[0])

    def _expand_path(self, path):
        """A helper to expand a path to a full absolute path."""
        return os.path.abspath(os.path.expanduser(path))


class EKSClient:
    def __init__(self, session, parsed_args, parsed_globals=None):
        self._session = session
        self._cluster_name = parsed_args.cluster_name
        self._cluster_description = None
        self._parsed_globals = parsed_globals
        self._parsed_args = parsed_args

    @property
    def cluster_description(self):
        """
        Use an eks describe-cluster call to get the cluster description
        Cache the response in self._cluster_description.
        describe-cluster will only be called once.
        """
        if self._cluster_description is not None:
            return self._cluster_description

        client_kwargs = {}
        if self._parsed_globals:
            client_kwargs.update(
                {
                    "region_name": self._parsed_globals.region,
                    "endpoint_url": self._parsed_globals.endpoint_url,
                    "verify": self._parsed_globals.verify_ssl,
                }
            )

        # Handle role assumption if needed
        if getattr(self._parsed_args, 'assume_role_arn', None):
            sts_client = self._session.create_client('sts')
            credentials = sts_client.assume_role(
                RoleArn=self._parsed_args.assume_role_arn,
                RoleSessionName='EKSDescribeClusterSession',
            )["Credentials"]

            client_kwargs.update(
                {
                    "aws_access_key_id": credentials["AccessKeyId"],
                    "aws_secret_access_key": credentials["SecretAccessKey"],
                    "aws_session_token": credentials["SessionToken"],
                }
            )

        client = self._session.create_client("eks", **client_kwargs)
        full_description = client.describe_cluster(name=self._cluster_name)
        cluster = full_description.get("cluster")

        if not cluster or "status" not in cluster:
            raise EKSClusterError("Cluster not found")
        if cluster["status"] not in ["ACTIVE", "UPDATING"]:
            raise EKSClusterError(f"Cluster status is {cluster['status']}")

        self._cluster_description = cluster
        return cluster

    def get_cluster_entry(self):
        """
        Return a cluster entry generated using
        the previously obtained description.
        """

        cert_data = self.cluster_description.get(
            "certificateAuthority", {}
        ).get("data", "")
        endpoint = self.cluster_description.get("endpoint")
        arn = self.cluster_description.get("arn")

        generated_cluster = OrderedDict(
            [
                (
                    "cluster",
                    OrderedDict(
                        [
                            ("certificate-authority-data", cert_data),
                            ("server", endpoint),
                        ]
                    ),
                ),
                ("name", arn),
            ]
        )

        if self._parsed_args.proxy_url is not None:
            generated_cluster["cluster"]["proxy-url"] = (
                self._parsed_args.proxy_url
            )

        return generated_cluster

    def get_user_entry(self, user_alias=None):
        """
        Return a user entry generated using
        the previously obtained description.
        """
        region = self.cluster_description.get("arn").split(":")[3]
        outpost_config = self.cluster_description.get("outpostConfig")

        if outpost_config is None:
            cluster_identification_parameter = "--cluster-name"
            cluster_identification_value = self._cluster_name
        else:
            # If cluster contains outpostConfig, use id for identification
            cluster_identification_parameter = "--cluster-id"
            cluster_identification_value = self.cluster_description.get("id")

        generated_user = OrderedDict(
            [
                (
                    "name",
                    user_alias or self.cluster_description.get("arn", ""),
                ),
                (
                    "user",
                    OrderedDict(
                        [
                            (
                                "exec",
                                OrderedDict(
                                    [
                                        ("apiVersion", API_VERSION),
                                        (
                                            "args",
                                            [
                                                "--region",
                                                region,
                                                "eks",
                                                "get-token",
                                                cluster_identification_parameter,
                                                cluster_identification_value,
                                                "--output",
                                                "json",
                                            ],
                                        ),
                                        ("command", "aws"),
                                    ]
                                ),
                            )
                        ]
                    ),
                ),
            ]
        )

        if self._parsed_args.role_arn is not None:
            generated_user["user"]["exec"]["args"].extend(
                ["--role", self._parsed_args.role_arn]
            )

        if self._session.profile:
            generated_user["user"]["exec"]["env"] = [
                OrderedDict(
                    [("name", "AWS_PROFILE"), ("value", self._session.profile)]
                )
            ]

        return generated_user



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from awscli.customizations.emr import hbase, ssh
from awscli.customizations.emr.addinstancegroups import AddInstanceGroups
from awscli.customizations.emr.addsteps import AddSteps
from awscli.customizations.emr.addtags import modify_tags_argument
from awscli.customizations.emr.command import override_args_required_option
from awscli.customizations.emr.createcluster import CreateCluster
from awscli.customizations.emr.createdefaultroles import CreateDefaultRoles
from awscli.customizations.emr.describecluster import DescribeCluster
from awscli.customizations.emr.installapplications import InstallApplications
from awscli.customizations.emr.listclusters import (
    modify_list_clusters_argument,
)
from awscli.customizations.emr.modifyclusterattributes import ModifyClusterAttr
from awscli.customizations.emr.terminateclusters import TerminateClusters


def emr_initialize(cli):
    """
    The entry point for EMR high level commands.
    """
    cli.register('building-command-table.emr', register_commands)
    cli.register('building-argument-table.emr.add-tags', modify_tags_argument)
    cli.register(
        'building-argument-table.emr.list-clusters',
        modify_list_clusters_argument,
    )
    cli.register(
        'before-building-argument-table-parser.emr.*',
        override_args_required_option,
    )


def register_commands(command_table, session, **kwargs):
    """
    Called when the EMR command table is being built. Used to inject new
    high level commands into the command list. These high level commands
    must not collide with existing low-level API call names.
    """
    command_table['terminate-clusters'] = TerminateClusters(session)
    command_table['describe-cluster'] = DescribeCluster(session)
    command_table['modify-cluster-attributes'] = ModifyClusterAttr(session)
    command_table['install-applications'] = InstallApplications(session)
    command_table['create-cluster'] = CreateCluster(session)
    command_table['add-steps'] = AddSteps(session)
    command_table['restore-from-hbase-backup'] = hbase.RestoreFromHBaseBackup(
        session
    )
    command_table['create-hbase-backup'] = hbase.CreateHBaseBackup(session)
    command_table['schedule-hbase-backup'] = hbase.ScheduleHBaseBackup(session)
    command_table['disable-hbase-backups'] = hbase.DisableHBaseBackups(session)
    command_table['create-default-roles'] = CreateDefaultRoles(session)
    command_table['add-instance-groups'] = AddInstanceGroups(session)
    command_table['ssh'] = ssh.SSH(session)
    command_table['socks'] = ssh.Socks(session)
    command_table['get'] = ssh.Get(session)
    command_table['put'] = ssh.Put(session)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


class AutoCompleter:
    """Main auto-completer object for the AWS CLI.

    This object delegates to concrete completers that can perform
    completions for specific cases (e.g model-based completions,
    server-side completions, etc).
    """

    def __init__(self, parser, completers):
        """

        :param parser: A parser.CLIParser instance.
        :param completers: A list of ``BaseCompleter`` instances.

        """
        self._parser = parser
        self._completers = completers

    def autocomplete(self, command_line, index=None):
        """Attempt to find completion suggestions.

        :param command_line: The currently entered command line as a string.
        :param index: An optional integer that indicates the location where
            the cursor is located (0 based index).

        :return: A list of ``CompletionResult`` objects.

        """
        parsed = self._parser.parse(command_line, index)
        for completer in self._completers:
            result = completer.complete(parsed)
            if result is not None:
                return result
        return []


class CompletionResult:
    """A data object for a single completion result.

    In addition to storing the completion string, this object also
    stores metadata about the completion.

    """

    def __init__(
        self,
        name,
        starting_index=0,
        required=False,
        cli_type_name='',
        help_text='',
        display_text=None,
    ):
        self.name = name
        self.starting_index = starting_index
        self.required = required
        self.cli_type_name = cli_type_name
        self.help_text = help_text
        self.display_text = display_text

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__)
            and self.name == other.name
            and self.starting_index == other.starting_index
            and self.display_text == other.display_text
        )

    def __repr__(self):
        return '%s(%s, %s, %s, %s, %s, %s)' % (
            self.__class__.__name__,
            self.name,
            self.starting_index,
            self.required,
            self.cli_type_name,
            self.help_text,
            self.display_text,
        )


class BaseCompleter:
    def complete(self, parsed):
        """Attempt to autocomplete parsed on parsed result.

        Subclasses should implement this method.

        :param parsed: A ParsedResult from the auto complete CLI parser.
        :return: This method should return one of two values:

            * ``None`` - If the completer doesn't understand how to
              complete the command, it should return ``None``.  This
              signals to the ``AutoCompleter`` that it should move on
              to the next completer.
            * ``List[CompletionResult]`` - If the completer is able to offer
              auto-completions it should return a list of strings that
              are valid suggestions for completing the command.  This
              indicates to the ``AutoCompleter`` to immediately return
              and to stop consulting other completers for results.
        """
        raise NotImplementedError("complete")



# Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import copy
from dataclasses import dataclass
from typing import Optional

from awscli.arguments import CLIArgument, CustomArgument
from awscli.customizations.exceptions import ParamValidationError


@dataclass
class ArgumentParameters:
    name: str
    member: Optional[str] = None
    help_text: Optional[str] = None
    required: Optional[bool] = False


class InjectingArgument(CustomArgument):
    def __init__(self, serialized_name, original_member_name, **kwargs):
        self._serialized_name = serialized_name
        self._original_member_name = original_member_name
        super().__init__(**kwargs)

    def add_to_params(self, parameters, value):
        if value is None:
            pass
        wrapped_value = {self._original_member_name: value}
        if parameters.get(self._serialized_name):
            parameters[self._serialized_name].update(wrapped_value)
        else:
            parameters[self._serialized_name] = wrapped_value


class OriginalArgument(CLIArgument):
    def __init__(self, original_member_name, error_message, **kwargs):
        self._serialized_name = kwargs.get("serialized_name")
        self._original_member_name = original_member_name
        self._error_message = error_message
        super().__init__(**kwargs)

    def add_to_params(self, parameters, value):
        if value is None:
            return

        unpacked = self._unpack_argument(value)
        if self._original_member_name in unpacked and self._error_message:
            raise ParamValidationError(self._error_message)

        if parameters.get(self._serialized_name):
            parameters[self._serialized_name].update(unpacked)
        else:
            parameters[self._serialized_name] = unpacked


class BinaryBlobArgumentHoister:
    def __init__(
        self,
        new_argument: ArgumentParameters,
        original_argument: ArgumentParameters,
        error_if_original_used: Optional[str] = None,
    ):
        self._new_argument = new_argument
        self._original_argument = original_argument
        self._error_message = error_if_original_used

    def __call__(self, session, argument_table, **kwargs):
        argument = argument_table[self._original_argument.name]
        model = copy.deepcopy(argument.argument_model)
        del model.members[self._original_argument.member]

        argument_table[self._new_argument.name] = InjectingArgument(
            argument._serialized_name,
            self._original_argument.member,
            name=self._new_argument.name,
            help_text=self._new_argument.help_text,
            cli_type_name="blob",
            required=self._new_argument.required,
        )
        argument_table[self._original_argument.name] = OriginalArgument(
            self._original_argument.member,
            self._error_message,
            name=self._original_argument.name,
            argument_model=model,
            operation_model=argument._operation_model,
            is_required=self._original_argument.required,
            event_emitter=session.get_component("event_emitter"),
            serialized_name=argument._serialized_name,
        )



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os

from prompt_toolkit.buffer import Buffer
from prompt_toolkit.document import Document
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.key_binding.bindings.focus import focus_next
from prompt_toolkit.keys import Keys
from prompt_toolkit.layout import Float, FloatContainer, HSplit, VSplit, Window
from prompt_toolkit.layout.controls import BufferControl
from prompt_toolkit.layout.dimension import Dimension
from prompt_toolkit.layout.layout import ConditionalContainer, Layout
from prompt_toolkit.layout.menus import (
    CompletionsMenu,
    MultiColumnCompletionsMenu,
)
from prompt_toolkit.layout.processors import BeforeInput
from prompt_toolkit.widgets import SearchToolbar, VerticalLine

from awscli.autoprompt.filters import (
    doc_section_visible,
    doc_window_has_focus,
    input_buffer_has_focus,
    is_history_mode,
    is_multi_column,
    is_one_column,
    output_section_visible,
)
from awscli.autoprompt.history import HistoryCompleter, HistoryDriver
from awscli.autoprompt.widgets import (
    DebugPanelWidget,
    HelpPanelWidget,
    TitleLine,
    ToolbarWidget,
)


class PrompterKeyboardInterrupt(KeyboardInterrupt):
    pass


class CLIPromptBuffer(Buffer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._completer = self.completer
        self._history_completer = HistoryCompleter(self)
        self.history_mode = False

    def switch_history_mode(self):
        self.history_mode = not self.history_mode
        if self.history_mode:
            self.completer = self._history_completer
            self.start_completion()
        else:
            self.completer = self._completer


class PromptToolkitFactory:
    DIMENSIONS = {
        'doc_window_height_max': 40,
        'doc_window_height_pref': 30,
        'input_buffer_height_min': 10,
        'menu_height_max': 16,
        'menu_scroll_offset': 2,
    }

    def __init__(self, completer, history_driver=None):
        self._completer = completer
        self._history_driver = history_driver

    @property
    def history_driver(self):
        if self._history_driver is None:
            cache_dir = os.path.expanduser(
                os.path.join('~', '.aws', 'cli', 'cache')
            )
            history_filename = os.path.join(cache_dir, 'prompt_history.json')
            self._history_driver = HistoryDriver(history_filename)
        return self._history_driver

    def create_input_buffer(self, on_text_changed_callback=None):
        return CLIPromptBuffer(
            name='input_buffer',
            completer=self._completer,
            history=self.history_driver,
            complete_while_typing=True,
            on_text_changed=on_text_changed_callback,
        )

    def create_doc_buffer(self):
        return Buffer(name='doc_buffer', read_only=True)

    def create_output_buffer(self):
        return Buffer(name='output_buffer', read_only=True)

    def create_input_buffer_container(self, input_buffer):
        return FloatContainer(
            Window(
                BufferControl(
                    buffer=input_buffer,
                    input_processors=[BeforeInput('> aws ')],
                ),
                height=Dimension(
                    min=self.DIMENSIONS['input_buffer_height_min']
                ),
                wrap_lines=True,
            ),
            [
                Float(
                    xcursor=True,
                    ycursor=True,
                    content=MultiColumnCompletionsMenu(
                        extra_filter=is_multi_column
                    ),
                ),
                Float(
                    xcursor=True,
                    ycursor=True,
                    content=CompletionsMenu(
                        extra_filter=is_one_column,
                        max_height=self.DIMENSIONS['menu_height_max'],
                        scroll_offset=self.DIMENSIONS['menu_scroll_offset'],
                    ),
                ),
            ],
        )

    def create_bottom_panel(self, doc_window, output_window):
        return VSplit(
            [
                ConditionalContainer(doc_window, doc_section_visible),
                ConditionalContainer(
                    VerticalLine(),
                    output_section_visible & doc_section_visible,
                ),
                ConditionalContainer(output_window, output_section_visible),
            ]
        )

    def create_searchable_window(self, title, output_buffer):
        search_field = SearchToolbar()
        return HSplit(
            [
                TitleLine(title),
                Window(
                    content=BufferControl(
                        buffer=output_buffer,
                        search_buffer_control=search_field.control,
                    ),
                    height=Dimension(
                        max=self.DIMENSIONS['doc_window_height_max'],
                        preferred=self.DIMENSIONS['doc_window_height_pref'],
                    ),
                    wrap_lines=True,
                ),
                search_field,
            ]
        )

    def create_layout(
        self,
        on_input_buffer_text_changed=None,
        input_buffer_container=None,
        doc_window=None,
        output_window=None,
    ):
        # This is the main layout, which consists of:
        # - The main input buffer with completion menus floating on top of it.
        # - A separating line between the input buffer and the doc window.
        # - A doc window to hold documentation.
        # - A separating line between the doc window and the toolbar.
        # - A toolbar denoting key bindings.
        # - A help panel
        # - A debug panel in case debug mode enabled
        if input_buffer_container is None:
            input_buffer = self.create_input_buffer(
                on_input_buffer_text_changed
            )
            input_buffer_container = self.create_input_buffer_container(
                input_buffer
            )
        if doc_window is None:
            doc_buffer = self.create_doc_buffer()
            doc_window = self.create_searchable_window('Doc panel', doc_buffer)
        if output_window is None:
            output_buffer = self.create_output_buffer()
            output_window = self.create_searchable_window(
                'Output panel', output_buffer
            )
        bottom_panel = self.create_bottom_panel(doc_window, output_window)
        return Layout(
            HSplit(
                [
                    VSplit(
                        [
                            HSplit([input_buffer_container, bottom_panel]),
                            HelpPanelWidget(),
                            DebugPanelWidget(),
                        ]
                    ),
                    ToolbarWidget(),
                ]
            )
        )

    def create_key_bindings(self):
        return PromptToolkitKeyBindings()


class PromptToolkitKeyBindings:
    def __init__(self, keybindings=None):
        if keybindings is None:
            keybindings = KeyBindings()
        self._kb = keybindings

        @self._kb.add(Keys.Enter, filter=input_buffer_has_focus)
        def _(event):
            buffer = event.app.current_buffer
            is_completing = getattr(buffer, 'complete_state', False)
            current_document = buffer.document
            if not is_completing:
                buffer.append_to_history()
                event.app.exit()
            else:
                # I didn't find better way to make a choice and close prompter
                # than reset buffer and change it to selected part
                if buffer.history_mode:
                    buffer.switch_history_mode()
                buffer.cancel_completion()
                event.app.current_buffer.reset()
                updated_document = Document(
                    text=current_document.text,
                    cursor_position=current_document.cursor_position,
                )
                buffer.set_document(updated_document)
                # If prompter suggested us something ended with slash and
                # started with 'file://' or 'fileb://' it should be path ended
                # with directory then we run completion again
                cur_word = current_document.get_word_under_cursor(WORD=True)
                if cur_word.endswith(os.sep) and cur_word.startswith(
                    ('file://', 'fileb://')
                ):
                    buffer.start_completion()

        @self._kb.add(Keys.Escape, filter=is_history_mode)
        def _(event):
            buffer = event.app.current_buffer
            buffer.cancel_completion()
            buffer.switch_history_mode()

        @self._kb.add(' ', filter=is_history_mode)
        def _(event):
            """Exit from history mode if something was selected or
            just add space to the end of the text and keep suggesting"""
            buffer = event.app.current_buffer
            if (
                buffer.complete_state
                and buffer.complete_state.current_completion
            ):
                buffer.switch_history_mode()
            buffer.insert_text(' ')

        @self._kb.add(Keys.F3)
        def _(event):
            current_buffer = event.app.current_buffer
            if current_buffer.name == 'doc_buffer':
                layout = event.app.layout
                input_buffer = layout.get_buffer_by_name('input_buffer')
                layout.focus(input_buffer)
            event.app.show_doc = not getattr(event.app, 'show_doc')

        @self._kb.add(Keys.F4)
        def _(event):
            event.app.multi_column = not event.app.multi_column

        @self._kb.add(Keys.F5)
        def _(event):
            event.app.show_output = not event.app.show_output
            if event.app.current_buffer.name == 'output_buffer':
                layout = event.app.layout
                input_buffer = layout.get_buffer_by_name('input_buffer')
                layout.focus(input_buffer)

        @self._kb.add(Keys.F1)
        def _(event):
            event.app.show_help = not event.app.show_help

        @self._kb.add(Keys.ControlR, filter=input_buffer_has_focus)
        def _(event):
            buffer = event.app.current_buffer
            buffer.cancel_completion()
            buffer.switch_history_mode()

        @self._kb.add(Keys.ControlC)
        @self._kb.add(Keys.ControlD)
        def _(event):
            layout = event.app.layout
            input_buffer = layout.get_buffer_by_name('input_buffer')
            text = f'> aws {input_buffer.document.text}'
            event.app.exit(exception=PrompterKeyboardInterrupt(text))

        @self._kb.add(Keys.F2)
        def _(event):
            focus_next(event)

        @self._kb.add('q', filter=~input_buffer_has_focus)
        def _(event):
            layout = event.app.layout
            input_buffer = layout.get_buffer_by_name('input_buffer')
            layout.focus(input_buffer)

        @self._kb.add('w', filter=doc_window_has_focus)
        def _(event):
            # Scroll up one unit equal to the height of the doc window.
            # Note: The scroll isn't exactly equal to the height of the doc
            # window. As there doesn't appear to be a way to get the exact
            # height of the doc window, the next best we can do is to grab the
            # preferred height of the Window.
            window_height = event.app.layout.current_window.height.preferred
            event.app.current_buffer.cursor_up(window_height)

        @self._kb.add('z', filter=doc_window_has_focus)
        def _(event):
            # Scroll down one unit equal to the height of the doc window.
            # Note: The scroll isn't exactly equal to the height of the doc
            # window. As there doesn't appear to be a way to get the exact
            # height of the doc window, the next best we can do is to grab the
            # preferred height of the Window.
            window_height = event.app.layout.current_window.height.preferred
            event.app.current_buffer.cursor_down(window_height)

        @self._kb.add('k', filter=doc_window_has_focus)
        def _(event):
            # Scroll up one line in the doc window.
            event.app.layout.current_buffer.cursor_up(1)

        @self._kb.add('j', filter=doc_window_has_focus)
        def _(event):
            # Scroll down one line in the doc window.
            event.app.layout.current_buffer.cursor_down(1)

        @self._kb.add('g', filter=doc_window_has_focus)
        def _(event):
            # Go to top of the doc window.
            document = event.app.layout.current_buffer.document
            current_row = document.cursor_position_row
            # `cursor_up() throws an error if its input is < 1`
            if current_row >= 1:
                event.app.layout.current_buffer.cursor_up(current_row)

        @self._kb.add('G', filter=doc_window_has_focus)
        def _(event):
            # Go to bottom of the doc window.
            document = event.app.layout.current_buffer.document
            current_row = document.cursor_position_row
            num_rows = document.line_count
            lines_to_bottom_delta = num_rows - current_row
            event.app.layout.current_buffer.cursor_down(lines_to_bottom_delta)

    @property
    def keybindings(self):
        return self._kb



# Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging
import sys

from awscli.customizations.commands import BasicCommand
from awscli.customizations.ec2instanceconnect.eicefetcher import (
    InstanceConnectEndpointRequestFetcher,
)
from awscli.customizations.ec2instanceconnect.eicesigner import (
    InstanceConnectEndpointRequestSigner,
)
from awscli.customizations.ec2instanceconnect.websocket import WebsocketManager
from awscli.customizations.exceptions import ParamValidationError

logger = logging.getLogger(__name__)


class OpenTunnelCommand(BasicCommand):
    NAME = "open-tunnel"

    DESCRIPTION = (
        "Opens a websocket tunnel to the specified EC2 Instance or private ip."
    )

    ARG_TABLE = [
        {
            "name": "instance-id",
            "help_text": "Specify the id of the EC2 instance that a tunnel should be opened for.",
            "required": False,
        },
        {
            "name": "instance-connect-endpoint-id",
            "help_text": "Specify the EC2 Instance Connect Endpoint Id that should be used to open the tunnel.",
            "required": False,
        },
        {
            "name": "instance-connect-endpoint-dns-name",
            "help_text": (
                "Specify the EC2 Instance Connect Endpoint DNS Name that should be used to open the tunnel. "
                "If this is specified, you must specify instance-connect-endpoint-id."
            ),
            "required": False,
        },
        {
            "name": "private-ip-address",
            "help_text": (
                "Specify the private ip address to open a tunnel for. "
                "If this is specified, you must specify instance-connect-endpoint-id."
            ),
            "required": False,
        },
        {
            "name": "remote-port",
            "help_text": "Specify the remote port to connect to on the instance. This defaults to port 22.",
            "default": 22,
            "cli_type_name": "integer",
            "required": False,
        },
        {
            "name": "local-port",
            "help_text": (
                "Specify the local port to listen on. Each new connection will have a separate websocket "
                "tunnel read and write to it."
            ),
            "cli_type_name": "integer",
            "required": False,
        },
        {
            "name": "max-tunnel-duration",
            "help_text": (
                "Specify the maximum time, in seconds, to keep a websocket tunnel alive for. This "
                "defaults to 3600 seconds."
            ),
            "cli_type_name": "integer",
            "required": False,
        },
        {
            "name": "max-websocket-connections",
            "help_text": (
                "Specify the maximum amount of websocket connections allowed when local-port is specified. This value "
                "defaults to 10. "
            ),
            "default": 10,
            "cli_type_name": "integer",
            "required": False,
        },
    ]

    def _run_main(self, parsed_args, parsed_globals):
        self._validate_parsed_args(parsed_args)
        ec2_client = self._session.create_client(
            "ec2",
            region_name=parsed_globals.region,
            verify=parsed_globals.verify_ssl,
        )

        instance_vpc_id = None
        instance_subnet_id = None
        private_ip_address = parsed_args.private_ip_address
        if not private_ip_address:
            instance_metadata = ec2_client.describe_instances(
                InstanceIds=[parsed_args.instance_id]
            )["Reservations"][0]["Instances"][0]
            instance_vpc_id = instance_metadata["VpcId"]
            instance_subnet_id = instance_metadata["SubnetId"]
            private_ip_address = instance_metadata["PrivateIpAddress"]

        instance_connect_endpoint_id = parsed_args.instance_connect_endpoint_id
        instance_connect_endpoint_dns_name = (
            parsed_args.instance_connect_endpoint_dns_name
        )
        if not instance_connect_endpoint_dns_name:
            eice_fetcher = InstanceConnectEndpointRequestFetcher()
            eice = eice_fetcher.get_available_instance_connect_endpoint(
                ec2_client,
                instance_vpc_id,
                instance_subnet_id,
                parsed_args.instance_connect_endpoint_id,
            )
            instance_connect_endpoint_id = eice["InstanceConnectEndpointId"]

            is_fips_enabled = self._session.get_config_variable(
                'use_fips_endpoint'
            )
            instance_connect_endpoint_dns_name = (
                eice_fetcher.get_eice_dns_name(eice, is_fips_enabled)
            )

        logger.debug(
            f"Using endpoint dns: {instance_connect_endpoint_dns_name}"
        )
        eice_request_signer = InstanceConnectEndpointRequestSigner(
            self._session,
            instance_connect_endpoint_dns_name=instance_connect_endpoint_dns_name,
            instance_connect_endpoint_id=instance_connect_endpoint_id,
            remote_port=parsed_args.remote_port,
            instance_private_ip=private_ip_address,
            max_tunnel_duration=parsed_args.max_tunnel_duration,
        )

        with WebsocketManager(
            parsed_args.local_port,
            parsed_args.max_websocket_connections,
            eice_request_signer,
            self._session.user_agent(),
        ) as websocket_manager:
            websocket_manager.run()
        return 0

    def _validate_parsed_args(self, parsed_args):
        if not parsed_args.instance_id and not parsed_args.private_ip_address:
            raise ParamValidationError("Specify an instance id or private ip.")
        if (
            parsed_args.instance_connect_endpoint_dns_name
            and not parsed_args.instance_connect_endpoint_id
        ):
            raise ParamValidationError(
                "Specify an instance connect endpoint id when providing a DNS name."
            )
        if (
            parsed_args.private_ip_address
            and not parsed_args.instance_connect_endpoint_id
        ):
            raise ParamValidationError(
                "Specify an instance connect endpoint id when providing a private ip."
            )
        if parsed_args.max_tunnel_duration is not None and (
            parsed_args.max_tunnel_duration < 1
            or parsed_args.max_tunnel_duration > 3_600
        ):
            raise ParamValidationError(
                "Invalid max connection timeout specified. Value must be greater than 1 and "
                "less than 3600."
            )
        if parsed_args.local_port is None and sys.stdin.isatty():
            raise ParamValidationError(
                "This command does not support interactive mode. You must use this command as a proxy or in listener "
                "mode. "
            )



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import ctypes
import os
import subprocess

DEFAULT_CONFIG_FILE = 'codedeploy.onpremises.yml'


class System:
    UNSUPPORTED_SYSTEM_MSG = (
        'Only Ubuntu Server, Red Hat Enterprise Linux Server and '
        'Windows Server operating systems are supported.'
    )

    def __init__(self, params):
        self.session = params.session
        self.s3 = self.session.create_client('s3', region_name=params.region)

    def validate_administrator(self):
        raise NotImplementedError('validate_administrator')

    def install(self, params):
        raise NotImplementedError('install')

    def uninstall(self, params):
        raise NotImplementedError('uninstall')


class Windows(System):
    CONFIG_DIR = r'C:\ProgramData\Amazon\CodeDeploy'
    CONFIG_FILE = 'conf.onpremises.yml'
    CONFIG_PATH = rf'{CONFIG_DIR}\{CONFIG_FILE}'
    INSTALLER = 'codedeploy-agent.msi'

    def validate_administrator(self):
        if not ctypes.windll.shell32.IsUserAnAdmin():
            raise RuntimeError(
                'You must run this command as an Administrator.'
            )

    def install(self, params):
        if 'installer' in params:
            self.INSTALLER = params.installer

        process = subprocess.Popen(
            [
                'powershell.exe',
                '-Command',
                'Stop-Service',
                '-Name',
                'codedeployagent',
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        (output, error) = process.communicate()
        not_found = (
            "Cannot find any service with service name 'codedeployagent'"
        )
        if process.returncode != 0 and not_found not in error:
            raise RuntimeError(
                f'Failed to stop the AWS CodeDeploy Agent:\n{error}'
            )

        response = self.s3.get_object(Bucket=params.bucket, Key=params.key)
        with open(self.INSTALLER, 'wb') as f:
            f.write(response['Body'].read())

        subprocess.check_call(
            [
                rf'.\{self.INSTALLER}',
                '/quiet',
                '/l',
                r'.\codedeploy-agent-install-log.txt',
            ],
            shell=True,
        )
        subprocess.check_call(
            [
                'powershell.exe',
                '-Command',
                'Restart-Service',
                '-Name',
                'codedeployagent',
            ]
        )

        process = subprocess.Popen(
            [
                'powershell.exe',
                '-Command',
                'Get-Service',
                '-Name',
                'codedeployagent',
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        (output, error) = process.communicate()
        if "Running" not in output:
            raise RuntimeError(
                'The AWS CodeDeploy Agent did not start after installation.'
            )

    def uninstall(self, params):
        process = subprocess.Popen(
            [
                'powershell.exe',
                '-Command',
                'Stop-Service',
                '-Name',
                'codedeployagent',
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        (output, error) = process.communicate()
        not_found = (
            "Cannot find any service with service name 'codedeployagent'"
        )
        if process.returncode == 0:
            self._remove_agent()
        elif not_found not in error:
            raise RuntimeError(
                f'Failed to stop the AWS CodeDeploy Agent:\n{error}'
            )

    def _remove_agent(self):
        process = subprocess.Popen(
            [
                'wmic',
                'product',
                'where',
                'name="CodeDeploy Host Agent"',
                'call',
                'uninstall',
                '/nointeractive',
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        (output, error) = process.communicate()
        if process.returncode != 0:
            raise RuntimeError(
                f'Failed to uninstall the AWS CodeDeploy Agent:\n{error}'
            )


class Linux(System):
    CONFIG_DIR = '/etc/codedeploy-agent/conf'
    CONFIG_FILE = DEFAULT_CONFIG_FILE
    CONFIG_PATH = f'{CONFIG_DIR}/{CONFIG_FILE}'
    INSTALLER = 'install'

    def validate_administrator(self):
        if os.geteuid() != 0:
            raise RuntimeError('You must run this command as sudo.')

    def install(self, params):
        if 'installer' in params:
            self.INSTALLER = params.installer

        self._update_system(params)
        self._stop_agent(params)

        response = self.s3.get_object(Bucket=params.bucket, Key=params.key)
        with open(self.INSTALLER, 'wb') as f:
            f.write(response['Body'].read())

        subprocess.check_call(['chmod', '+x', f'./{self.INSTALLER}'])

        credentials = self.session.get_credentials()
        environment = os.environ.copy()
        environment['AWS_REGION'] = params.region
        environment['AWS_ACCESS_KEY_ID'] = credentials.access_key
        environment['AWS_SECRET_ACCESS_KEY'] = credentials.secret_key
        if credentials.token is not None:
            environment['AWS_SESSION_TOKEN'] = credentials.token
        subprocess.check_call([f'./{self.INSTALLER}', 'auto'], env=environment)

    def uninstall(self, params):
        process = self._stop_agent(params)
        if process.returncode == 0:
            self._remove_agent(params)

    def _update_system(self, params):
        raise NotImplementedError('preinstall')

    def _remove_agent(self, params):
        raise NotImplementedError('remove_agent')

    def _stop_agent(self, params):
        process = subprocess.Popen(
            ['service', 'codedeploy-agent', 'stop'],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        (output, error) = process.communicate()
        if process.returncode != 0 and params.not_found_msg not in error:
            raise RuntimeError(
                f'Failed to stop the AWS CodeDeploy Agent:\n{error}'
            )
        return process


class Ubuntu(Linux):
    def _update_system(self, params):
        subprocess.check_call(['apt-get', '-y', 'update'])
        subprocess.check_call(['apt-get', '-y', 'install', 'ruby2.0'])

    def _remove_agent(self, params):
        subprocess.check_call(['dpkg', '-r', 'codedeploy-agent'])

    def _stop_agent(self, params):
        params.not_found_msg = 'codedeploy-agent: unrecognized service'
        return Linux._stop_agent(self, params)


class RHEL(Linux):
    def _update_system(self, params):
        subprocess.check_call(['yum', '-y', 'install', 'ruby'])

    def _remove_agent(self, params):
        subprocess.check_call(['yum', '-y', 'erase', 'codedeploy-agent'])

    def _stop_agent(self, params):
        params.not_found_msg = (
            'Redirecting to /bin/systemctl stop  codedeploy-agent.service'
        )
        return Linux._stop_agent(self, params)
